import os
import yaml
from copy import deepcopy
import logging
from aequilibrae.context import get_active_project


class Parameters:
    """Global parameters module

    Parameters are used in many procedures, and are often defined only in the parameters.yml file ONLY
    Parameters are organized in the following groups:

    * assignment
    * distribution
    * system
    * report zeros
    * temp directory

    .. code-block:: python

        >>> from aequilibrae import Project, Parameters

        >>> project = Project()
        >>> project.new(tmp_path_empty)

        >>> p = Parameters(project)

        >>> p.parameters['system']['logging_directory'] =  "/tmp/other_folder"
        >>> p.parameters['osm']['overpass_endpoint'] = "http://192.168.0.110:32780/api"
        >>> p.parameters['osm']['max_query_area_size'] = 10000000000
        >>> p.parameters['osm']['sleeptime'] = 0
        >>> p.write_back()

        >>> # You can also restore the software default values
        >>> p.restore_default()
    """

    _default: dict
    file_default: str

    def __init__(self, project=None):
        """Loads parameters from file. The place is always the same. The root of the package"""
        project = project or get_active_project(must_exist=False)
        proj_path = project.project_base_path if project is not None else ""

        self.file = os.path.join(proj_path, "parameters.yml")

        if os.path.isfile(self.file):
            with open(self.file, "r") as yml:
                self.parameters = yaml.load(yml, Loader=yaml.SafeLoader)
        else:
            if project is not None:
                logger = logging.getLogger("aequilibrae")
                logger.warning("No pre-existing parameter file exists for this project. Will use default")

            self.parameters = deepcopy(self._default)

    def write_back(self):
        """Writes the parameters back to file"""
        with open(self.file, "w") as stream:
            yaml.dump(self.parameters, stream, default_flow_style=False)

    def restore_default(self):
        """Restores parameters to generic default"""
        self.parameters = self._default
        self.write_back()


Parameters.file_default = os.path.join(os.path.dirname(os.path.realpath(__file__)), "parameters.yml")
with open(Parameters.file_default, "r") as yml:
    Parameters._default = yaml.safe_load(yml)



import pandas as pd
import numpy as np
cimport numpy as np
cimport cython

from libcpp.queue cimport queue

@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.initializedcheck(False)
cdef void _remove_dead_ends(
    long long [:] graph_fs,
    long long [:] all_nodes,
    long long [:] nodes_to_indices,
    long long [:] a_nodes,
    long long [:] b_nodes,
    signed char [:] directions,
    long long [:] in_degree,
    long long [:] out_degree,
    np.uint8_t [:] burnt_links,
) noexcept nogil:
    cdef:
        long long b_node
        Py_ssize_t node_idx, incoming, outgoing

        queue[long long] Q

    # Discovery: we are looking for potential dead ends, this includes:
    #     - nodes with all incoming edges,
    #     - nodes with all outgoing edges,
    #     - nodes for which all out going links point to the same node and has the same number of back links.
    #       This is a generalisation of dead-ends formed by a node with a single bidirectional link to another
    #
    # Removal: At nodes of interest we begin a Breadth First Search (BFS) for links we can remove. It's uncommon for the BFS to extend far as the stopping
    # criteria and conditions for expansion are very similar, often this just searches along a line.
    # For removal the node must have "no choices", that is, all outgoing edges point to the same node *and* all incoming edges can be account for as
    # from that same node.
    # The criteria for expansion is that there are no incoming edges *and* all outgoing edges point to the same node.
    for starting_node_idx in range(all_nodes.shape[0]):
        node = all_nodes[starting_node_idx]

        Q.push(starting_node_idx)
        while not Q.empty():
            node_idx = Q.front()
            Q.pop()

            # Centroids are marked with negative degrees, the actual degree does not matter
            if in_degree[node_idx] < 0 or out_degree[node_idx] < 0:
                continue
            elif in_degree[node_idx] == 0 and out_degree[node_idx] == 0:
                continue
            elif in_degree[node_idx] > 0 and out_degree[node_idx] == 0:
                # All incoming or all outgoing edges, since there's no way to either leave, or get to this node all the attached
                # edges are of no use. However we have no (current) means to remove these edges, a transpose of the graph would be required to
                # avoid individual lookups.
                continue

            ### Expansion
            if in_degree[node_idx] == 0 and out_degree[node_idx] > 0:
                for link_idx in range(graph_fs[node_idx], graph_fs[node_idx + 1]):
                    if not burnt_links[link_idx]:
                        # Only need to burn links leaving this node
                        burnt_links[link_idx] = True
                        in_degree[b_nodes[link_idx]] -= 1
                        out_degree[node_idx] -= 1
                        Q.push(b_nodes[link_idx])
                continue

            ### Propagation
            # We now know that the node we are looking at has a mix of incoming and outgoing edges, i.e. in_degree[node] > 0 and out_degree[node] > 0
            # That implies that this node is reachable from some other node. We now need to assess if this node would ever be considered in pathfinding.
            # To be considered, there needs to be some form of real choice, i.e. there needs to be multiple ways to leave a new node. If the only way
            # to leave this node is back the way we came then the cost of the path
            #     ... -> pre -> node -> pre -> ...
            # is greater than that of
            #     ... -> pre -> ...
            # so we can remove this link to node. This is because negative cost cycles are disallowed in path finding.
            #
            # We don't remove the case where there is only one new node that could be used to leave as it is handled in the graph compression.
            # It would however, be possible to handle that here as well.

            # If all the outgoing edges are to the same node *and* all incoming edges come from that node, then there is no real choice and the
            # link should be removed. This is primarily to catch nodes who are only connected to a single other node via a bidirectional link but can
            # handle the case where multiple links point between the same pair of nodes with perhaps, different, but still non-negative costs.

            # Lets first find a link that isn't burnt. Then check that every link after this one points to the same node
            b_node = -1
            for link_idx in range(graph_fs[node_idx], graph_fs[node_idx + 1]):
                if burnt_links[link_idx]:
                    continue

                if b_node == -1:
                    b_node = b_nodes[link_idx]
                elif b_node != b_nodes[link_idx]:
                    # We've found a link to a different node, that means we have some other way to leave and we can't remove this node.
                    break
            else:
                # We now know all outgoing edges point to the same node. Lets now check that all incoming edges are accounted for
                incoming = in_degree[node_idx]
                for link_idx in range(graph_fs[b_node], graph_fs[b_node + 1]):
                    # Incoming degree has already been decremented, when a link was burnt
                    if not burnt_links[link_idx] and b_nodes[link_idx] == node_idx:
                        incoming -= 1

                if incoming != 0:
                    # Not all incoming edges are accounted for, there is some other node that links to this one. From where we don't know.
                    continue

                # All incoming edges are accounted for, this node is of interest.
                for link_idx in range(graph_fs[node_idx], graph_fs[node_idx + 1]):
                    if not burnt_links[link_idx]:
                        burnt_links[link_idx] = True
                        in_degree[b_node] -= 1
                        out_degree[node_idx] -= 1

                for link_idx in range(graph_fs[b_node], graph_fs[b_node + 1]):
                    if not burnt_links[link_idx] and b_nodes[link_idx] == node_idx:
                        burnt_links[link_idx] = True
                        in_degree[node_idx] -= 1
                        out_degree[b_node] -= 1

                Q.push(b_node)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef long long _build_compressed_graph(long long[:] link_idx,
                                  long long[:] links_index,
                                  long long[:] link_edge,
                                  long long[:] a_nodes,
                                  long long[:] b_nodes,
                                  signed char[:] directions,
                                  long long link_id_max,
                                  long long[:] simplified_links,
                                  signed char[:] simplified_directions,
                                  long long[:] counts,
                                  long long[:] all_links,
                                  long long[:] compressed_dir,
                                  long long[:] compressed_a_node,
                                  long long[:] compressed_b_node) noexcept nogil:
    cdef:
        long long slink = 0
        long long pre_link, n, first_node, lnk, lidx, a_node, b_node
        bint ab_dir, ba_dir
        long drc
        Py_ssize_t i, k

    # For each link we have marked for examining
    for pre_link in link_edge:
        if simplified_links[pre_link] >= 0:
            continue

        # We grab the initial information for that link
        lidx = link_idx[pre_link]
        a_node = a_nodes[lidx]
        b_node = b_nodes[lidx]
        drc = directions[lidx]

        n = a_node if counts[a_node] == 2 else b_node
        first_node = b_node if counts[a_node] == 2 else a_node

        # True if there exists a link in the direction ab (or ba), could be two links, or a single bidirectional link
        ab_dir = False if (first_node == a_node and drc < 0) or (first_node == b_node and drc > 0) else True
        ba_dir = False if (first_node == a_node and drc > 0) or (first_node == b_node and drc < 0) else True

        # While the node we are looking at, n, has degree two, we can continue to compress
        while counts[n] == 2:
            simplified_links[pre_link] = slink  # Mark link for removal
            simplified_directions[pre_link] = -1 if a_node == n else 1

            # Gets the link from the list that is not the link we are coming from
            for k in range(links_index[n], links_index[n + 1]):
                if pre_link != all_links[k]:
                    pre_link = all_links[k]
                    break

            lidx = link_idx[pre_link]
            a_node = a_nodes[lidx]
            b_node = b_nodes[lidx]
            drc = directions[lidx]
            ab_dir = False if (n == a_node and drc < 0) or (n == b_node and drc > 0) else ab_dir
            ba_dir = False if (n == a_node and drc > 0) or (n == b_node and drc < 0) else ba_dir
            n = (
                a_nodes[lidx]
                if n == b_nodes[lidx]
                else b_nodes[lidx]
            )

        simplified_links[pre_link] = slink
        simplified_directions[pre_link] = -1 if a_node == n else 1
        last_node = b_node if counts[a_node] == 2 else a_node
        # major_nodes[slink] = [first_node, last_node]

        # Available directions are NOT indexed like the other arrays
        compressed_a_node[slink] = first_node
        compressed_b_node[slink] = last_node
        if ab_dir:
            if ba_dir:
                compressed_dir[slink] = 0
            else:
                compressed_dir[slink] = 1
        elif ba_dir:
            compressed_dir[slink] = -1
        else:
            compressed_dir[slink] = -999
        slink += 1

    return slink


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef void _back_fill(long long[:] links_index, long long max_node) noexcept:
    cdef Py_ssize_t i

    for i in range(max_node + 1, 0, -1):
        links_index[i - 1] = links_index[i] if links_index[i - 1] == -1 else links_index[i - 1]


def build_compressed_graph(graph):
    # General notes:
    # Anything that uses graph.network is operating on the **mixed** graph. This graph has both directed and undirected edges
    # Anything that uses graph.graph is operating on the **directed** graph. This graph has only directed edges, they may be backwards but they are directed

    burnt_links = np.full(len(graph.graph), False, dtype=bool)

    directed_node_max = max(graph.graph.a_node.values.max(), graph.graph.b_node.values.max())
    in_degree = np.bincount(graph.graph.b_node.values, minlength=directed_node_max + 1)
    out_degree = np.bincount(graph.graph.a_node.values, minlength=directed_node_max + 1)

    centroid_idx = graph.nodes_to_indices[graph.centroids]
    in_degree[centroid_idx] = -1
    out_degree[centroid_idx] = -1
    del centroid_idx

    _remove_dead_ends(
        graph.fs,
        graph.all_nodes,
        graph.nodes_to_indices,
        graph.graph.a_node.values,
        graph.graph.b_node.values,
        graph.graph.direction.values,
        in_degree,
        out_degree,
        burnt_links,
    )

    graph.dead_end_links = graph.graph.link_id.values[burnt_links]  # Perhaps filter to unique link_ids? There'll be duplicates in here
    df = pd.DataFrame(graph.network, copy=True)
    if graph.dead_end_links.shape[0]:
        df = df[~df.link_id.isin(graph.dead_end_links)]

    # Build link index
    link_id_max = df.link_id.max()

    link_idx = np.empty(link_id_max + 1, dtype=np.int64)
    link_idx[df.link_id] = np.arange(df.shape[0])

    nodes = np.hstack([df.a_node.values, df.b_node.values])
    links = np.hstack([df.link_id.values, df.link_id.values])
    counts = np.bincount(nodes)  # index (node) i has frequency counts[i]. This is just the number of edges that connect to a given node

    idx = np.argsort(nodes)
    all_nodes = nodes[idx]
    all_links = links[idx]
    all_nodes_max = all_nodes.max()

    links_index = np.full(all_nodes_max + 2, -1, dtype=np.int64)
    nlist = np.arange(all_nodes_max + 2)

    y, x, _ = np.intersect1d(all_nodes, nlist, assume_unique=False, return_indices=True)
    links_index[y] = x[:]
    links_index[-1] = all_links.shape[0]

    _back_fill(links_index[:], all_nodes_max)

    # We keep all centroids for sure
    counts[graph.centroids] = 999

    degree_two = (counts == 2).astype(np.uint8)
    # Reorder and sum the degree two nodes by how they appear in the network, finds how a particular node is connected, resulting values are
    # 0: This node is not of degree one. We're not interested in this case.
    # 1: This node is of degree one AND, has either incoming or outgoing flow. We can remove these these.
    link_edge = df.link_id.values[
        degree_two[df.a_node.values] + degree_two[df.b_node.values] == 1
    ].astype(np.int64)

    simplified_links = np.full(link_id_max + 1, -1, dtype=np.int64)
    simplified_directions = np.zeros(link_id_max + 1, dtype=np.int8)

    compressed_dir = np.zeros(link_id_max + 1, dtype=np.int64)
    compressed_a_node = np.zeros(link_id_max + 1, dtype=np.int64)
    compressed_b_node = np.zeros(link_id_max + 1, dtype=np.int64)

    slink = _build_compressed_graph(
        link_idx[:],
        links_index[:],
        link_edge[:],
        df.a_node.values[:],
        df.b_node.values[:],
        df.direction.values[:],
        link_id_max,
        simplified_links[:],
        simplified_directions[:],
        counts[:],
        all_links[:],
        compressed_dir[:],
        compressed_a_node[:],
        compressed_b_node[:]
    )

    links_to_remove = (simplified_links >= 0).nonzero()[0]
    if links_to_remove.shape[0]:
        df = df[~df.link_id.isin(links_to_remove)]
        df = df[df.a_node != df.b_node]

    comp_lnk = pd.DataFrame(
        {
            "a_node": compressed_a_node[:slink],
            "b_node": compressed_b_node[:slink],
            "direction": compressed_dir[:slink],
            "link_id": np.arange(slink),
        }
    )

    # Link compression can introduce new simple cycles into the graph
    comp_lnk = comp_lnk[comp_lnk.a_node != comp_lnk.b_node]

    max_link_id = link_id_max * 10
    comp_lnk.link_id += max_link_id

    df = pd.concat([df, comp_lnk])
    df = df[["id", "link_id", "a_node", "b_node", "direction"]]
    properties = graph._build_directed_graph(df, graph.centroids)
    graph.compact_all_nodes = properties[0]
    graph.compact_num_nodes = properties[1]
    graph.compact_nodes_to_indices = properties[2]
    graph.compact_fs = properties[3]
    graph.compact_graph = properties[4]

    crosswalk = pd.DataFrame(
        {
            "link_id": np.arange(simplified_directions.shape[0]),
            "link_direction": simplified_directions,
            "compressed_link": simplified_links,
            "compressed_direction": np.ones(simplified_directions.shape[0], dtype=np.int8),
        }
    )

    crosswalk = crosswalk[crosswalk.compressed_link >= 0]
    crosswalk.compressed_link += max_link_id

    cw2 = pd.DataFrame(crosswalk, copy=True)
    cw2.link_direction *= -1
    cw2.compressed_direction = -1

    crosswalk = pd.concat([crosswalk, cw2])
    crosswalk = crosswalk.assign(key=crosswalk.compressed_link * crosswalk.compressed_direction)
    crosswalk.drop(["compressed_link", "compressed_direction"], axis=1, inplace=True)

    final_ids = pd.DataFrame(graph.compact_graph[["id", "link_id", "direction"]], copy=True)
    final_ids = final_ids.assign(key=final_ids.link_id * final_ids.direction)
    final_ids.drop(["link_id", "direction"], axis=1, inplace=True)

    agg_crosswalk = crosswalk.merge(final_ids, on="key")
    agg_crosswalk.loc[:, "key"] = agg_crosswalk.link_id * agg_crosswalk.link_direction
    agg_crosswalk.drop(["link_id", "link_direction"], axis=1, inplace=True)

    direct_crosswalk = final_ids[final_ids.key.abs() < max_link_id]

    crosswalk = pd.concat([agg_crosswalk, direct_crosswalk])[["key", "id"]]
    crosswalk.columns = ["__graph_correlation_key__", "__compressed_id__"]

    graph.graph = graph.graph.assign(__graph_correlation_key__=graph.graph.link_id * graph.graph.direction)
    graph.graph = graph.graph.merge(crosswalk, on="__graph_correlation_key__", how="left")
    graph.graph.drop(["__graph_correlation_key__"], axis=1, inplace=True)

    # If will refer all the links that have no correlation to an element beyond the last link
    # This element will always be zero during assignment
    graph.graph.__compressed_id__ = graph.graph.__compressed_id__.fillna(graph.compact_graph.id.max() + 1).astype(np.int64)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.initializedcheck(False)
def create_compressed_link_network_mapping(graph):
        # Cache the result, this isn't a huge computation but isn't worth doing twice
        if (
            graph.compressed_link_network_mapping_idx is not None
            and graph.compressed_link_network_mapping_data is not None
            and graph.network_compressed_node_mapping is not None
        ):
            return (
                graph.compressed_link_network_mapping_idx,
                graph.compressed_link_network_mapping_data,
                graph.network_compressed_node_mapping,
            )

        cdef:
            long long i, j, a_node, x, b_node, tmp, compressed_id
            long long[:] b
            long long[:] values
            np.uint32_t[:] idx
            np.uint32_t[:] data
            np.int32_t[:] node_mapping

        # This method requires that graph.graph is sorted on the a_node IDs, since that's done already we don't
        # bother redoing sorting it.

        # Some links are completely removed from the network, they are assigned ID `graph.compact_graph.id.max() + 1`,
        # we skip them.
        filtered = graph.graph[graph.graph.__compressed_id__ != graph.compact_graph.id.max() + 1]
        gb = filtered.groupby(by="__compressed_id__", sort=True)
        idx = np.zeros(graph.compact_num_links + 1, dtype=np.uint32)
        data = np.zeros(len(filtered), dtype=np.uint32)

        node_mapping = np.full(graph.num_nodes, -1, dtype=np.int32)

        i = 0
        for compressed_id, df in gb:
            idx[compressed_id] = i
            values = df.link_id.values
            a = df.a_node.values
            b = df.b_node.values

            # In order to ensure that the link IDs come out in the correct order we must walk the links
            # we do this assuming the `a` array is sorted.
            j = 0
            # Find the missing a_node, this is the starting of the chain. We cannot rely on the node ordering to do a simple lookup

            a_node = x = a[np.isin(a, b, invert=True, assume_unique=True)][0]
            while True:
                tmp = a.searchsorted(x)
                if tmp < len(a) and a[tmp] == x:
                    x = b[tmp]
                    data[i + j] = values[tmp]
                else:
                    break
                j += 1

            b_node = x
            node_mapping[a_node] = graph.compact_graph["a_node"].iat[compressed_id]
            node_mapping[b_node] = graph.compact_graph["b_node"].iat[compressed_id]

            i += len(values)

        idx[-1] = i

        graph.compressed_link_network_mapping_idx = np.array(idx)
        graph.compressed_link_network_mapping_data = np.array(data)
        graph.network_compressed_node_mapping = np.array(node_mapping)

        return idx, data, node_mapping



from aequilibrae.project.database_connection import database_connection
from aequilibrae.utils.db_utils import commit_and_close


def delete_pattern(pattern_id: int):
    """Deletes all information regarding one specific transit_pattern.

    :Arguments:
        **pattern_id** (:obj:`str`): pattern_id as present in the database
    """
    sqls = [
        """DELETE from trips where trip_id IN
                (select trip_id from trips where pattern_id=?)""",
        "DELETE from trips where pattern_id=?",
        "DELETE from links where pattern_id=?",
        "DELETE from pattern_links where pattern_id=?",
        "DELETE from pattern_mapping where pattern_id=?",
    ]

    with commit_and_close(database_connection("transit")) as conn:
        for sql in sqls:
            conn.execute(sql, [pattern_id])



import numpy as np
import pandas as pd
import logging
from scipy.sparse import coo_matrix as sp_coo_matrix

cdef class GeneralisedCOODemand:
    def __init__(self, origin_col: str, destination_col: str, nodes_to_indices, shape=None):
        """
        A container for access to the float64 and float32 fields of a data frame.
        """
        self.df = pd.DataFrame(columns=[origin_col, destination_col]).set_index([origin_col, destination_col])
        self.shape = shape
        self.nodes_to_indices = nodes_to_indices

    def add_df(self, dfs: Union[pd.DataFrame, List[pd.DataFrame]], shape=None, fill: float = 0.0):
        """
        Add a DataFrame to the existing ones.

        Expects a DataFrame with a multi-index of (o, d).
        """
        if isinstance(dfs, pd.DataFrame):
            dfs = (dfs,)

        if shape is None and self.shape is None:
            raise ValueError("a shape must be provided initially to prevent oddly sized sparse matrices")
        if shape is not None and self.shape is None:
            self.shape = shape
        if shape is not None and self.shape is not None and shape != self.shape:
            raise ValueError(f"provided shape ({shape}) differs from previous shape ({self.shape})")

        new_dfs = [self.df]
        for df in dfs:
            if df.index.nlevels != 2:
                raise ValueError("provided pd.DataFrame doesn't have a 2-level multi-index")
            elif df.index.names != self.df.index.names:
                raise ValueError(f"mismatched index names. Expect {self.df.index.names}, provided {df.index.names}")

            shape = self.nodes_to_indices[df.index.to_frame(index=False)].max(axis=0)
            if shape[0] >= self.shape[0] or shape[1] >= self.shape[1]:
                raise ValueError(f"inferred max index ({(shape[0], shape[1])}) exceeds provided shape ({self.shape})")

            new_dfs.append(df.select_dtypes(["float64", "float32"]))

        self.df = pd.concat(new_dfs, axis=1).fillna(fill).sort_index()

    def add_matrix(self, matrix: AequilibraeMatrix, shape=None, fill: float = 0.0):
        """
        Add an AequilibraE matrix to the existing demand in a sparse manner.
        """
        dfs = []
        for i, name in enumerate(matrix.view_names):
            assert name not in self.df.columns, f"Matrix name ({name}) already exists in the matrix cube"
            m = matrix.matrix_view if len(matrix.view_names) == 1 else matrix.matrix_view[:, :, i]
            if np.nansum(m) == 0:
                continue

            coo_ = sp_coo_matrix(m)
            df = pd.DataFrame(
                data={
                    self.df.index.names[0]: matrix.index[coo_.row],
                    self.df.index.names[1]: matrix.index[coo_.col],
                    name: coo_.data,
                },
            ).set_index([self.df.index.names[0], self.df.index.names[1]])
            dfs.append(df.dropna())

        self.add_df(dfs, shape=shape, fill=fill)
        logging.info(f"There {len(self.df):,} are OD pairs with non-zero flows")

    def _initalise_c_data(self):
        self.ods = self.df.index  # MultiIndex[int, int] -> vector[pair[long long, long long]]

        self.f64.clear()
        self.f32.clear()
        self.f64_names = []
        self.f32_names = []

        cdef:
            double[::1] f64_array
            float[::1] f32_array
            vector[double] *f64_vec
            vector[float] *f32_vec

        for col in self.df:
            if self.df.dtypes[col] == "float64":
                f64_array = self.df[col].to_numpy()
                self.f64_names.append(col)

                # The unique pointer will take ownership of this allocation
                f64_vec = new vector[double]()
                f64_vec.insert(f64_vec.begin(), &f64_array[0], &f64_array[0] + len(f64_array))
                self.f64.emplace_back(f64_vec)  # From here f63_vec should not be accessed. It is owned by the unique pointer

            elif self.df.dtypes[col] == "float32":
                f32_array = self.df[col].to_numpy()
                self.f32_names.append(col)

                # The unique pointer will take ownership of this allocation
                f32_vec = new vector[float]()
                f32_vec.insert(f32_vec.begin(), &f32_array[0], &f32_array[0] + len(f32_array))
                self.f32.emplace_back(f32_vec)  # From here f32_vec should not be accessed. It is owned by the unique pointer
            else:
                raise TypeError(f"non-floating point column ({col}) in self.df. Something has gone wrong")

    def no_demand(GeneralisedCOODemand self) -> bool:
        return len(self.df.columns) == 0

    def is_empty(self) -> bool:
        return self.df.index.empty



"""
=============
path computation related code
=============

STILL NEED TO ADD SOME EXPLANATIONS HERE

"""

__author__ = "Pedro Camargo ($Author: Pedro Camargo $)"
__version__ = "1.0"
__revision__ = "$Revision: 1 $"
__date__ = "$Date: 2016-07-02$"

from .assignment_results import AssignmentResults, TransitAssignmentResults
from .path_results import PathResults
from .skim_results import SkimResults



import pandas as pd


def get_table(table_name, conn):
    """
    Selects table from database.

    :Arguments:
         **table_name** (:obj:`str`): desired table name
         **conn** (:obj:`sqlite3.Connection`): database connection
    """

    return pd.read_sql(f"SELECT * FROM {table_name};", con=conn)



from aequilibrae.project.project import Project
from aequilibrae.project.about import About
from aequilibrae.project.network.network import Network
from aequilibrae.project.field_editor import FieldEditor
from aequilibrae.project.zoning import Zoning
from aequilibrae.project.zone import Zone
from aequilibrae.project.data import Matrices
from aequilibrae.log import Log
from aequilibrae.project.network.period import Period
from aequilibrae.project.network.periods import Periods



import csv
from os.path import join

import pandas as pd

# from polarislib.network.data import DataTableStorage
from aequilibrae.utils.get_table import get_table


def write_fares(folder_path: str, conn):
    fattr = get_table("Transit_Fare_Attributes", conn).reset_index()
    fattr.rename(columns={"currency": "currency_type", "transfer": "transfers"}, inplace=True)
    fattr.transfer_duration = fattr.transfer_duration.astype(int)

    headers = ["fare_id", "price", "currency_type", "payment_method", "transfers", "agency_id", "transfer_duration"]

    fattr[headers].to_csv(join(folder_path, "fare_attributes.txt"), quoting=csv.QUOTE_NONNUMERIC, index=False)

    frls = get_table("Transit_Fare_Rules", conn).reset_index()
    frls.rename(columns={fld: f"{fld}_id" for fld in frls.columns if "id" not in fld}, inplace=True)
    frls = frls[["fare_id", "route_id", "origin_id", "destination_id", "contains_id"]]

    for fld in ["origin_id", "destination_id"]:
        frls[fld].fillna(-99999, inplace=True)
        frls[fld] = frls[fld].astype(int).astype(str)
        frls.loc[frls[fld] == "-99999", fld] = pd.NA
    frls.to_csv(join(folder_path, "fare_rules.txt"), quoting=csv.QUOTE_NONNUMERIC, index=False)



import importlib.util as iutil
import warnings
from typing import Union, List

from shapely.geometry import Point, Polygon, LineString, MultiPoint, MultiPolygon, MultiLineString
from shapely.wkb import loads

rtree_avail = iutil.find_spec("rtree") is not None
qgis = iutil.find_spec("qgis") is not None
if qgis:
    from qgis.core import QgsSpatialIndex as Index
    from qgis.core import QgsGeometry, QgsFeature

    env = "QGIS"
elif rtree_avail:
    from rtree.index import Index as Index

    env = "Python"
else:
    env = "NOT AVAILABLE"


class GeoIndex:
    """Implements a generic GeoIndex class that uses the QGIS index when using the GUI and RTree otherwise"""

    def __init__(self):
        self.idx = Index()
        self.built = False

    def build_from_layer(self, layer) -> dict:
        if env != "QGIS":
            warnings.warn("This method works inside QGIS only")
        self.built = True
        self.idx = Index(layer.getFeatures())
        return {f.id(): loads(f.geometry().asWkb().data()) for f in layer.getFeatures()}

    def insert(
        self, feature_id: int, geometry: Union[Point, Polygon, LineString, MultiPoint, MultiPolygon, MultiLineString]
    ) -> None:
        """Inserts a valid shapely geometry in the index

        :Arguments:
            **feature_id** (:obj:`int`): ID of the geometry being inserted
            **geo** (:obj:`Shapely.geometry`): Any valid shapely geometry
        """
        self.built = True
        if env == "QGIS":
            g = QgsGeometry()
            g.fromWkb(geometry.wkb)
            feature = QgsFeature()
            feature.setGeometry(g)
            feature.setId(feature_id)
            self.idx.addFeature(feature)
        elif env == "Python":
            self.idx.insert(feature_id, geometry.bounds)
        else:
            warnings.warn("You need RTREE to build a spatial index")

    def nearest(self, geo: Union[Point, Polygon, LineString, MultiPoint, MultiPolygon], num_results) -> List[int]:
        """Finds nearest neighbor for a given geometry

        :Arguments:
            **geo** (:obj:`Shapely geometry`): Any valid shapely geometry

            **num_results** (:obj:`int`): A positive integer for the number of neighbors to return

        :Returns:
            **neighbors** (:obj:`List[int]`): List of IDs of the closest neighbors in the index
        """
        if env == "QGIS":
            g = QgsGeometry()
            g.fromWkb(geo.wkb)
            return self.idx.nearestNeighbor(g, num_results)
        elif env == "Python":
            return self.idx.nearest(geo.bounds, num_results)
        else:
            warnings.warn("You need RTREE to build a spatial index")

    def delete(self, feature_id, geometry: Union[Point, Polygon, LineString, MultiPoint, MultiPolygon]):
        if env == "QGIS":
            g = QgsGeometry()
            g.fromWkb(geometry.wkb)
            feature = QgsFeature()
            feature.setGeometry(g)
            feature.setId(feature_id)
            self.idx.deleteFeature(feature)
        elif env == "Python":
            self.idx.delete(feature_id, geometry.bounds)
        else:
            warnings.warn("You need RTREE to build a spatial index")

    def reset(self):
        self.idx = Index()
        self.built = False



import dataclasses
from sqlite3 import Connection
from typing import Dict, Any, Optional

from shapely.geometry import Point

from aequilibrae.transit.constants import Constants, AGENCY_MULTIPLIER
from aequilibrae.transit.transit_elements.basic_element import BasicPTElement


@dataclasses.dataclass
class Stop(BasicPTElement):
    """Transit stop as read from the GTFS feed"""

    def __init__(self, agency_id: int, record: tuple, headers: list):
        self.stop_id = -1
        self.stop = ""
        self.stop_code = ""
        self.stop_name = ""
        self.stop_desc = ""
        self.stop_lat: float = None
        self.stop_lon: float = None
        self.stop_street = ""
        self.zone = ""
        self.zone_id = None
        self.stop_url = ""
        self.location_type = 0
        self.parent_station = ""
        self.stop_timezone = ""

        # Not part of GTFS
        self.taz = None
        self.agency = ""
        self.agency_id = agency_id
        self.link = None
        self.dir = None
        self.srid = -1
        self.geo: Optional[Point] = None
        self.route_type: Optional[int] = None
        self.__map_matching_id__: Dict[Any, Any] = {}
        self.__moved_map_matching__ = 0

        for key, value in zip(headers, record):
            if key not in self.__dict__.keys():
                raise KeyError(f"{key} field in Stops.txt is unknown field for that file on GTFS")
            key = key if key != "stop_id" else "stop"
            key = key if key != "zone_id" else "zone"
            self.__dict__[key] = value

        if None not in [self.stop_lon, self.stop_lat]:
            self.geo = Point(self.stop_lon, self.stop_lat)
        if len(str(self.zone_id)) == 0:
            self.zone_id = None

    def save_to_database(self, conn: Connection, commit=True) -> None:
        """Saves Transit Stop to the database"""

        sql = """insert into stops (stop_id, stop, agency_id, link, dir, name,
                                    parent_station, description, street, fare_zone_id, transit_zone, route_type, geometry)
                 values (?,?,?,?,?,?,?,?,?,?,?,?, GeomFromWKB(?, ?));"""

        dt = self.data
        conn.execute(sql, dt)
        if commit:
            conn.commit()

    @property
    def data(self) -> list:
        return [
            self.stop_id,
            self.stop,
            self.agency_id,
            self.link,
            self.dir,
            self.stop_name,
            self.parent_station,
            self.stop_desc,
            self.stop_street,
            self.zone_id,
            self.taz,
            int(self.route_type),
            self.geo.wkb,
            self.srid,
        ]

    def get_node_id(self):
        c = Constants()

        val = 1 + c.stops.get(self.agency_id, AGENCY_MULTIPLIER * self.agency_id)
        c.stops[self.agency_id] = val
        self.stop_id = c.stops[self.agency_id]



import sqlite3

import shapely.wkb
import shapely.wkt
from shapely.ops import transform
from aequilibrae.utils.get_table import get_table

from aequilibrae.transit.transit_elements import Pattern


def read_patterns(conn: sqlite3.Connection, transformer):
    patterns = []
    data = get_table("transit_patterns", conn).reset_index()
    if not data.shape[0]:
        return
    data.geo = data.geo.apply(shapely.wkb.loads)

    data.drop(columns=["matching_quality"], inplace=True)
    data.rename(columns={"pattern": "pattern_hash", "geo": "shape"}, inplace=True)

    for idx, dt in data.iterrows():
        pat = Pattern(None, dt.route_id, None).from_row(dt)
        pat.shape_length = pat.best_shape().length

        if transformer:
            pat.shape = transform(transformer.transform, pat.shape)

        patterns.append(pat)
    return patterns



from math import radians, cos, sin, asin, sqrt


# from https://stackoverflow.com/a/4913653/1480643
def haversine(lon1, lat1, lon2, lat2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))
    r = 6371000  # Radius of earth in meters. Use 3956 for miles
    return c * r



""""
Large portions of this code were adopted from OSMNx, by Geoff Boeing.

Although attempts to use OSMNx were made (including refactoring its
entire code base as a contribution to that package), it became clear
that its integration with libraries not available with QGIS' Python
distribution was too tight, and was therefore not practical to
detach them in order to use OSMNx as a dependency or submodule

For the original work, please see https://github.com/gboeing/osmnx
"""

import gc
import logging
import re
import time
from typing import List, Dict

import pandas as pd
import requests
from shapely import Polygon

from aequilibrae.context import get_logger
from aequilibrae.parameters import Parameters
from aequilibrae.utils.signal import SIGNAL
from .osm_params import http_headers, memory


class OSMDownloader:
    downloading = SIGNAL(object)

    def __emit_all(self, *args):
        self.downloading.emit(*args)

    def __init__(self, polygons: List[Polygon], modes, logger: logging.Logger = None):
        self.logger = logger or get_logger()
        self.polygons = polygons
        self.filter = self.get_osm_filter(modes)
        self.report = []
        self.json = []
        par = Parameters().parameters["osm"]
        self.overpass_endpoint = par["overpass_endpoint"]
        self.timeout = par["timeout"]
        self.sleeptime = par["sleeptime"]
        self._nodes = []
        self._links = []
        self.data: Dict[str, pd.DataFrame] = {"nodes": pd.DataFrame([]), "links": pd.DataFrame([])}

    def doWork(self):
        infrastructure = 'way["highway"]'
        query_template = (
            "{memory}[out:json][timeout:{timeout}];({infrastructure}{filters}({south:.6f},{west:.6f},"
            "{north:.6f},{east:.6f});>;);out;"
        )
        self.__emit_all(["maxValue", len(self.polygons)])
        self.__emit_all(["Value", 0])
        m = ""
        if memory > 0:
            m = f"[maxsize: {memory}]"
        for counter, poly in enumerate(self.polygons):
            msg = f"Downloading polygon {counter + 1} of {len(self.polygons)}"
            self.logger.info(msg)
            self.__emit_all(["Value", counter])
            self.__emit_all(["text", msg])
            west, south, east, north = poly.bounds
            query_str = query_template.format(
                north=north,
                south=south,
                east=east,
                west=west,
                infrastructure=infrastructure,
                filters=self.filter,
                timeout=self.timeout,
                memory=m,
            )
            json = self.overpass_request(data={"data": query_str}, timeout=self.timeout)
            if json["elements"]:
                for tag, lst in [("node", self._nodes), ("way", self._links)]:
                    df = pd.DataFrame([item for item in json["elements"] if item["type"] == tag])
                    if tag == "node":
                        df = df.assign(is_centroid=0, modes="", link_types="", node_id=0)
                    lst.append(df)
                del json
                gc.collect()

        self.__emit_all(["Value", len(self.polygons)])
        self.__emit_all(["text", "Downloading finished. Processing data"])
        for lst, table in [(self._links, "links"), (self._nodes, "nodes")]:
            df = pd.DataFrame([])
            if len(lst) > 0:
                df = pd.concat(lst, ignore_index=True).drop_duplicates(subset=["id"]).drop(columns=["type"])
            if table != "links":
                df = df.drop(columns=["tags"], errors="ignore")
            self.data[table] = df.rename(columns={"id": "osm_id"}, errors="ignore")
            lst.clear()
            gc.collect()

        self.__emit_all(["FinishedDownloading", 0])

    def overpass_request(self, data, pause_duration=None, timeout=180, error_pause_duration=None):
        """Send a request to the Overpass API via HTTP POST and return the JSON response.

        :Arguments:
            **data**(:obj:`dict` or `OrderedDict`): key-value pairs of parameters to post to the API
            **pause_duration** (:obj:`int`): how long to pause in seconds before requests, if None, will query API
            status endpoint to find when next slot is available
            **timeout** (:obj:`int`): the timeout interval for the requests library
            **error_pause_duration**(:obj:`int`): how long to pause in seconds before re-trying requests if error

        :Returns:
            :obj:`dict`
        """

        # define the Overpass API URL, then construct a GET-style URL as a string to
        url = self.overpass_endpoint.rstrip("/") + "/interpreter"
        if pause_duration is None:
            time.sleep(self.sleeptime)
        start_time = time.time()
        self.report.append(f'Posting to {url} with timeout={timeout}, "{data}"')
        self.logger.debug(f'Posting to {url} with timeout={timeout}, "{data}"')
        response = requests.post(url, data=data, timeout=timeout, headers=http_headers)

        # get the response size and the domain, log result
        size_kb = len(response.content) / 1000.0
        domain = re.findall(r"(?s)//(.*?)/", url)[0]
        msg = "Downloaded {:,.1f}KB from {} in {:,.2f} seconds".format(size_kb, domain, time.time() - start_time)
        self.report.append(msg)
        self.logger.info(msg)

        try:
            response_json = response.json()
            if "remark" in response_json:
                msg = f'Server remark: "{response_json["remark"]}"'
                self.report.append(msg)
                self.logger.info(msg)
        except Exception as err:
            # 429 is 'too many requests' and 504 is 'gateway timeout' from server
            # overload - handle these errors by recursively calling
            # overpass_request until we get a valid response
            if response.status_code in [429, 504]:
                # pause for error_pause_duration seconds before re-trying request
                if error_pause_duration is None:
                    error_pause_duration = self.sleeptime + 1
                msg = "Server at {} returned status code {} and no JSON data. Re-trying request in {:.2f} seconds.".format(
                    domain, response.status_code, error_pause_duration
                )
                self.report.append(msg)
                self.logger.info(msg)
                time.sleep(error_pause_duration)
                response_json = self.overpass_request(data=data, pause_duration=pause_duration, timeout=timeout)

            # else, this was an unhandled status_code, throw an exception
            else:
                self.report.append(f"Server at {domain} returned status code {response.status_code} and no JSON data")
                raise Exception(
                    f"Server returned no JSON data.\n{response} {response.reason}\n{response.text}"
                ) from err

        return response_json

    def get_osm_filter(self, modes: list) -> str:
        """
        loosely adapted from http://www.github.com/gboeing/osmnx
        """

        p = Parameters().parameters["network"]["osm"]
        all_tags = p["all_link_types"]

        p = p["modes"]
        all_modes = list(p.keys())

        tags_to_keep = []
        for m in modes:
            if m not in all_modes:
                raise ValueError(f"Mode {m} not listed in the parameters file")
            tags_to_keep += p[m]["link_types"]
        tags_to_keep = list(set(tags_to_keep))

        # Default to remove
        service = '["service"!~"parking|parking_aisle|driveway|private|emergency_access"]'
        access = '["access"!~"private"]'

        filtered = [x for x in all_tags if x not in tags_to_keep]
        filtered = "|".join(filtered)

        filter = f'["area"!~"yes"]["highway"!~"{filtered}"]{service}{access}'

        return filter



import tempfile
import os
import glob
import sys
from aequilibrae.log import logger, global_logger
from aequilibrae.parameters import Parameters
from aequilibrae.project.data import Matrices
from aequilibrae.log import Log
from aequilibrae import matrix
from aequilibrae import transit
from aequilibrae import project

try:
    from aequilibrae.paths.AoN import path_computation
except Exception as e:
    global_logger.warning(f"Failed to import compiled modules. {e.args}")
    raise

from aequilibrae.distribution import Ipf, GravityApplication, GravityCalibration, SyntheticGravityModel
from aequilibrae.matrix import AequilibraeMatrix, AequilibraeData
from aequilibrae import distribution
from aequilibrae.paths.network_skimming import NetworkSkimming
from aequilibrae.paths.traffic_class import TrafficClass
from aequilibrae.paths.vdf import VDF
from aequilibrae.paths.all_or_nothing import allOrNothing
from aequilibrae.paths.traffic_assignment import TrafficAssignment
from aequilibrae.paths.graph import Graph
from aequilibrae.paths.assignment_paths import AssignmentPaths
from aequilibrae.project import Project
from aequilibrae.paths.results import AssignmentResults, SkimResults, PathResults

from aequilibrae import paths

name = "aequilibrae"


def setup():
    sys.dont_write_bytecode = True
    cleaning()


def cleaning():
    p = tempfile.gettempdir() + "/aequilibrae_*"
    for f in glob.glob(p):
        try:
            os.unlink(f)
        except Exception as err:
            global_logger.warning(err.__str__())


setup()



from .safe_class import SafeClass
from ...utils.db_utils import commit_and_close
from ...utils.spatialite_utils import connect_spatialite


class Period(SafeClass):
    """A Period object represents a single record in the *periods* table

    .. code-block:: python

        >>> from aequilibrae import Project

        >>> proj = Project.from_path("/tmp/test_project")

        >>> all_periods = proj.network.periods

        # We can just get one link in specific
        >>> period1 = all_periods.get(1)

        # We can find out which fields exist for the period
        >>> which_fields_do_we_have = period1.data_fields()

        # It succeeds if the period_id already does not exist
        >>> period1.renumber(998877)

        # We can just save the period
        >>> period1.save()
    """

    def __init__(self, dataset, project):
        """"""
        super().__init__(dataset, project)
        self.__fields = list(dataset.keys())
        self._table = "periods"

    def save(self):
        """Saves period to database"""

        with commit_and_close(connect_spatialite(self.project.path_to_file)) as conn:
            if self.period_id != self.__original__["period_id"]:
                raise ValueError("One cannot change the period_id")

            data, sql = self.__save_period()

            if data:
                conn.execute(sql, data)

    def data_fields(self) -> list:
        """Lists all data fields for the period, as available in the database

        :Returns:
            **data fields** (:obj:`list`): list of all fields available for editing
        """

        return list(self.__original__.keys())

    def renumber(self, new_id: int):
        """Renumbers the period in the network

        Logs a warning if another period already exists with this period_id

        :Arguments:
            **new_id** (:obj:`int`): New period_id
        """

        new_id = int(new_id)

        if new_id == 1 or self.period_id == 1:
            raise ValueError("You cannot renumber, or renumber another period to the default period.")

        if new_id == self.period_id:
            self._logger.warning("This is already the period number")
            return

        with commit_and_close(connect_spatialite(self.project.path_to_file)) as conn:
            try:
                conn.execute("Update periods set period_id=? where period_id=?", [new_id, self.period_id])
            finally:
                conn.commit()
        self._logger.info(f"Period {self.period_id} was renumbered to {new_id}")
        self.__dict__["period_id"] = new_id
        self.__original__["period_id"] = new_id

    def __save_period(self):
        data = []
        txts = []
        for key, val in self.__dict__.items():
            if key not in self.__original__:
                continue
            data.append(val)
            txts.append(f'"{key}"')

        if not data:
            self._logger.warning(f"Nothing to update for period {self.period_id}")
            return [], ""

        values = ",".join("?" * len(txts))
        txts = ",".join(txts)
        sql = f"INSERT OR REPLACE INTO periods ({txts}) VALUES ({values})"
        return data, sql

    def __setattr__(self, instance, value) -> None:
        if instance not in self.__dict__ and instance[:1] != "_":
            raise AttributeError(f'"{instance}" is not a valid attribute for a period')
        elif instance == "period_id":
            raise AttributeError("Setting period_id is not allowed")
        self.__dict__[instance] = value



import logging


_current_project = None


def activate_project(project):
    global _current_project
    _current_project = project


def get_active_project(must_exist=True):
    if not _current_project and must_exist:
        raise FileNotFoundError("There is no active Project set")
    return _current_project


def get_logger():
    proj = get_active_project(must_exist=False)
    if proj is None or proj.logger is None:
        return logging.getLogger("aequilibrae")
    return proj.logger



import importlib.util as iutil


def noop(_):
    pass


if iutil.find_spec("PyQt5") is not None:
    from PyQt5.QtCore import pyqtSignal as SIGNAL  # type: ignore

    noop(SIGNAL.__class__)  # This should be no-op but it stops PyCharm from "optimising" the above import
else:
    from aequilibrae.utils.python_signal import PythonSignal as SIGNAL  # type: ignore

    noop(SIGNAL.__class__)  # This should be no-op but it stops PyCharm from "optimising" the above import



import pickle
import uuid
from abc import ABC
from datetime import datetime
from os.path import join
from typing import List, Tuple, Optional
import dataclasses

import numpy as np
import pandas as pd
from aequilibrae.paths.graph_building import build_compressed_graph, create_compressed_link_network_mapping

from aequilibrae.context import get_logger


@dataclasses.dataclass
class NetworkGraphIndices:
    network_ab_idx: np.array
    network_ba_idx: np.array
    graph_ab_idx: np.array
    graph_ba_idx: np.array


def _get_graph_to_network_mapping(lids, direcs):
    num_uncompressed_links = int(np.unique(lids).shape[0])
    indexing = np.zeros(int(lids.max()) + 1, np.uint64)
    indexing[np.unique(lids)[:]] = np.arange(num_uncompressed_links)

    graph_ab_idx = direcs > 0
    graph_ba_idx = direcs < 0
    network_ab_idx = indexing[lids[graph_ab_idx]]
    network_ba_idx = indexing[lids[graph_ba_idx]]
    return NetworkGraphIndices(network_ab_idx, network_ba_idx, graph_ab_idx, graph_ba_idx)


class GraphBase(ABC):  # noqa: B024
    """
    Graph class.

    AequilibraE graphs implement two forms of compression.
        - link contraction, and
        - dead end removal.

    Link contraction creates a topological equivalent graph by contracting sequences of links between nodes
    with degrees of two. This compresses long streams of links, such as along highways or curved roads, into single links.

    Dead end removal attempts to remove dead ends and fish spines from the network. It does this based on the observation
    that in a graph with non-negative weights a dead end will only ever appear in the results of a short(est) path if the
    origin or destination is present within that dead end.

    Dead end removal is applied before link contraction and does not create a strictly topological equivalent graph,
    however, all centroids are preserved.

    The compressed graph is used internally.
    """

    def __init__(self, logger=None):
        self.logger = logger or get_logger()
        self.__integer_type = np.int64
        self.__float_type = np.float64

        self.required_default_fields = ["link_id", "a_node", "b_node", "direction", "id"]
        self.__required_default_types = [
            self.__integer_type,
            self.__integer_type,
            self.__integer_type,
            np.int8,
            self.__integer_type,
        ]
        self.other_fields = ""
        self.mode = ""
        self.date = str(datetime.now())

        self.description = "No description added so far"

        self.num_links = -1
        self.num_nodes = -1
        self.num_zones = -1

        self.compact_num_links = -1
        self.compact_num_nodes = -1

        self.network = pd.DataFrame([])  # This method will hold ALL information on the network
        self.graph = pd.DataFrame([])  # This method will hold an array with ALL fields in the graph.

        self.compact_graph = pd.DataFrame([])  # This method will hold an array with ALL fields in the graph.

        # These are the fields actually used in computing paths
        self.all_nodes = np.array(0)  # Holds an array with all nodes in the original network
        self.nodes_to_indices = np.array(0, np.int64)  # Holds the reverse of the all_nodes
        self.fs = np.array([])  # This method will hold the forward star for the graph
        self.cost = np.array([])  # This array holds the values being used in the shortest path routine
        self.skims = None

        self.lonlat_index = pd.DataFrame([])  # Holds a node_id to lon/lat coord index for nodes within this graph

        self.compact_all_nodes = np.array(0)  # Holds an array with all nodes in the original network
        self.compact_nodes_to_indices = np.array(0)  # Holds the reverse of the all_nodes
        self.compact_fs = np.array([])  # This method will hold the forward star for the graph
        self.compact_cost = np.array([])  # This array holds the values being used in the shortest path routine
        self.compact_skims = None

        self.capacity = np.array([])  # Array holds the capacity for links
        self.free_flow_time = np.array([])  # Array holds the free flow travel time by link

        # sake of the Cython code
        self.skim_fields = []  # List of skim fields to be used in computation
        self.cost_field = False  # Name of the cost field

        self.block_centroid_flows = True
        self.penalty_through_centroids = np.inf

        self.centroids = None  # NumPy array of centroid IDs

        self.g_link_crosswalk = np.array([])  # 4 a link ID in the BIG graph, a corresponding link in the compressed 1

        self.dead_end_links = np.array([])

        self.compressed_link_network_mapping_idx = None
        self.compressed_link_network_mapping_data = None

        # Randomly generate a unique Graph ID randomly
        self._id = uuid.uuid4().hex

    def default_types(self, tp: str):
        """
        Returns the default integer and float types used for computation

        :Arguments:
            **tp** (:obj:`str`): data type. 'int' or 'float'
        """
        if tp == "int":
            return self.__integer_type
        elif tp == "float":
            return self.__float_type
        else:
            raise ValueError("It must be either a int or a float")

    def prepare_graph(self, centroids: Optional[np.ndarray]) -> None:
        """
        Prepares the graph for a computation for a certain set of centroids

        Under the hood, if sets all centroids to have IDs from 1 through **n**,
        which should correspond to the index of the matrix being assigned.

        This is what enables having any node IDs as centroids, and it relies on
        the inference that all links connected to these nodes are centroid
        connectors.

        :Arguments:
            **centroids** (:obj:`np.ndarray`): Array with centroid IDs. Mandatory type Int64, unique and positive
        """
        self.__network_error_checking__()

        # Creates the centroids

        if centroids is not None:
            if not np.issubdtype(centroids.dtype, np.integer):
                raise ValueError("Centroids need to be a NumPy array of integers 64 bits")
            if centroids.shape[0] == 0:
                raise ValueError("You need at least one centroid")
            if centroids.min() <= 0:
                raise ValueError("Centroid IDs need to be positive")
            if centroids.shape[0] != np.unique(centroids).shape[0]:
                raise ValueError("Centroid IDs are not unique")
            self.centroids = np.array(centroids, np.uint32)
        else:
            self.centroids = np.array([], np.uint32)

        self.network = self.network.astype(
            {
                "direction": np.int8,
                "a_node": self.__integer_type,
                "b_node": self.__integer_type,
                "link_id": self.__integer_type,
            }
        )

        properties = self._build_directed_graph(self.network, self.centroids)
        self.all_nodes, self.num_nodes, self.nodes_to_indices, self.fs, self.graph = properties

        # We generate IDs that we KNOW will be constant across modes
        self.graph.sort_values(by=["link_id", "direction"], inplace=True)
        self.graph["__supernet_id__"] = np.arange(self.graph.shape[0]).astype(self.__integer_type)
        self.graph.sort_values(by=["a_node", "b_node"], inplace=True)

        self.num_links = self.graph.shape[0]
        self.__build_derived_properties()

        if self.centroids.shape[0]:
            self.__build_compressed_graph()
            self.compact_num_links = self.compact_graph.shape[0]

        # The cache property should be recalculated when the graph has been re-prepared
        self.compressed_link_network_mapping_idx = None
        self.compressed_link_network_mapping_data = None
        self.network_compressed_node_mapping = None

    def __build_compressed_graph(self):
        build_compressed_graph(self)

        # We build a groupby to save time later
        self.__graph_groupby = self.graph.groupby(["__compressed_id__"])

    def _build_directed_graph(self, network: pd.DataFrame, centroids: np.ndarray):
        all_titles = list(network.columns)

        not_pos = network.loc[network.direction != 1, :]
        not_negs = network.loc[network.direction != -1, :]

        names, types = self.__build_column_names(all_titles)
        neg_names = []
        for name in names:
            if name in not_pos.columns:
                neg_names.append(name)
            elif name + "_ba" in not_pos.columns:
                neg_names.append(name + "_ba")
        not_pos = pd.DataFrame(not_pos, copy=True)[neg_names]
        not_pos.columns = names

        # Swap the a and b nodes of these edges. Direction is used for mapping the graph.graph back to the network. It does not indicate the direction of the link.
        not_pos.loc[:, "direction"] = -1
        aux = np.array(not_pos.a_node.values, copy=True)
        not_pos.loc[:, "a_node"] = not_pos.loc[:, "b_node"]
        not_pos.loc[:, "b_node"] = aux[:]
        del aux

        pos_names = []
        for name in names:
            if name in not_negs.columns:
                pos_names.append(name)
            elif name + "_ab" in not_negs.columns:
                pos_names.append(name + "_ab")
        not_negs = pd.DataFrame(not_negs, copy=True)[pos_names]
        not_negs.columns = names
        not_negs.loc[:, "direction"] = 1

        df = pd.concat([not_negs, not_pos])

        # Now we take care of centroids
        nodes = np.unique(np.hstack((df.a_node.values, df.b_node.values))).astype(self.__integer_type)
        nodes = np.setdiff1d(nodes, centroids, assume_unique=True)
        all_nodes = np.hstack((centroids, nodes)).astype(self.__integer_type)

        num_nodes = all_nodes.shape[0]

        nodes_to_indices = np.full(int(all_nodes.max()) + 1, -1, dtype=np.int64)
        nlist = np.arange(num_nodes)
        nodes_to_indices[all_nodes] = nlist

        df.a_node = nodes_to_indices[df.a_node.values]
        df.b_node = nodes_to_indices[df.b_node.values]
        df = df.sort_values(by=["a_node", "b_node"])
        df.index = np.arange(df.shape[0])
        df["id"] = np.arange(df.shape[0])
        fs = np.empty(num_nodes + 1, dtype=self.__integer_type)
        fs.fill(-1)
        y, x, _ = np.intersect1d(df.a_node.values, nlist, assume_unique=False, return_indices=True)
        fs[y] = x[:]
        fs[-1] = df.shape[0]
        for i in range(num_nodes, 1, -1):
            if fs[i - 1] == -1:
                fs[i - 1] = fs[i]

        nans = ", ".join([i for i in df.columns if df[i].isnull().any().any()])
        if nans:
            self.logger.warning(f"Field(s) {nans} has(ve) at least one NaN value. Check your computations")

        df["link_id"] = df["link_id"].astype(self.__integer_type)
        df["b_node"] = df.b_node.values.astype(self.__integer_type)
        df["id"] = df.id.values.astype(self.__integer_type)
        df["direction"] = df.direction.values.astype(np.int8)

        return all_nodes, num_nodes, nodes_to_indices, fs, df

    def exclude_links(self, links: list) -> None:
        """
        Excludes a list of links from a graph by setting their B node equal to their A node

        :Arguments:
            **links** (:obj:`list`): List of link IDs to be excluded from the graph
        """
        filter = self.network.link_id.isin(links)
        # We check is the list makes sense in order to warn the user
        if filter.sum() != len(set(links)):
            self.logger.warning("At least one link does not exist in the network and therefore cannot be excluded")

        self.network.loc[filter, "b_node"] = self.network.loc[filter, "a_node"]

        if self.centroids is not None:
            self.prepare_graph(self.centroids)
            self.set_blocked_centroid_flows(self.block_centroid_flows)
        self._id = uuid.uuid4().hex

    def __build_column_names(self, all_titles: List[str]) -> Tuple[list, list]:
        fields = list(self.required_default_fields)
        types = list(self.__required_default_types)
        for column in all_titles:
            if column not in self.required_default_fields and column[0:-3] not in self.required_default_fields:
                if column[-3:] == "_ab":
                    if column[:-3] + "_ba" in all_titles:
                        fields.append(column[:-3])
                        types.append(self.network[column].dtype)
                    else:
                        raise ValueError("Field {} exists for ab direction but does not exist for ba".format(column))
                elif column[-3:] == "_ba":
                    if column[:-3] + "_ab" not in all_titles:
                        raise ValueError("Field {} exists for ba direction but does not exist for ab".format(column))
                else:
                    fields.append(column)
                    types.append(self.network[column].dtype)
        return fields, types

    def __build_dtype(self, all_titles) -> list:
        dtype = [
            ("link_id", self.__integer_type),
            ("a_node", self.__integer_type),
            ("b_node", self.__integer_type),
            ("direction", np.int8),
            ("id", self.__integer_type),
        ]
        for i in all_titles:
            if i not in self.required_default_fields and i[0:-3] not in self.required_default_fields:
                if i[-3:] == "_ab":
                    if i[:-3] + "_ba" in all_titles:
                        dtype.append((i[:-3], self.network[i].dtype))
                    else:
                        raise ValueError("Field {} exists for ab direction but does not exist for ba".format(i))
                elif i[-3:] == "_ba":
                    if i[:-3] + "_ab" not in all_titles:
                        raise ValueError("Field {} exists for ba direction but does not exist for ab".format(i))
                else:
                    dtype.append((i, self.network[i].dtype))
        return dtype

    def set_graph(self, cost_field) -> None:
        """
        Sets the field to be used for path computation

        :Arguments:
            **cost_field** (:obj:`str`): Field name. Must be numeric
        """
        if cost_field in self.graph.columns:
            self.cost_field = cost_field
            self.compact_cost = np.zeros(self.compact_graph.id.max() + 2, self.__float_type)
            df = self.__graph_groupby.sum(numeric_only=True)[[cost_field]].reset_index()
            self.compact_cost[df.index.values] = df[cost_field].values
            if self.graph[cost_field].dtype == self.__float_type:
                self.cost = np.array(self.graph[cost_field].values, copy=True)
            else:
                self.cost = np.array(self.graph[cost_field].values, dtype=self.__float_type)
                self.logger.warning("Cost field with wrong type. Converting to float64")
        else:
            raise ValueError("cost_field not available in the graph:" + str(self.graph.columns))

        self.__build_derived_properties()

    def set_skimming(self, skim_fields: list) -> None:
        """
        Sets the list of skims to be computed

        Skimming with A* may produce results that differ from traditional Dijkstra's due to its use a heuristic.

        :Arguments:
            **skim_fields** (:obj:`list`): Fields must be numeric
        """
        if not skim_fields:
            self.skim_fields = []
            self.skims = np.array([])

        if isinstance(skim_fields, str):
            skim_fields = [skim_fields]
        elif not isinstance(skim_fields, list):
            raise ValueError("You need to provide a list of skims or the same of a single field")

        # Check if list of fields make sense
        k = [x for x in skim_fields if x not in self.graph.columns]
        if k:
            raise ValueError("At least one of the skim fields does not exist in the graph: {}".format(",".join(k)))

        self.compact_skims = np.zeros((self.compact_num_links + 1, len(skim_fields) + 1), self.__float_type)
        df = self.__graph_groupby.sum(numeric_only=True)[skim_fields].reset_index()
        for i, skm in enumerate(skim_fields):
            self.compact_skims[df.index.values, i] = df[skm].values.astype(self.__float_type)

        self.skims = np.zeros((self.num_links, len(skim_fields) + 1), self.__float_type)
        t = [x for x in skim_fields if self.graph[x].dtype != self.__float_type]
        if t:
            Warning("Some skim field with wrong type. Converting to float64")
            for i, j in enumerate(skim_fields):
                self.skims[:, i] = self.graph[j].astype(self.__float_type).values[:]
        else:
            for i, j in enumerate(skim_fields):
                self.skims[:, i] = self.graph[j].values[:]
        self.skim_fields = skim_fields

    def set_blocked_centroid_flows(self, block_centroid_flows) -> None:
        """
        Chooses whether we want to block paths to go through centroids or not.

        Default value is ``True``

        :Arguments:
            **block_centroid_flows** (:obj:`bool`): Blocking or not
        """
        if not isinstance(block_centroid_flows, bool):
            raise TypeError("Blocking flows through centroids needs to be boolean")
        if self.num_zones == 0:
            self.logger.warning("No centroids in the model. Nothing to block")
            return
        self.block_centroid_flows = block_centroid_flows

    # Procedure to pickle graph and save to disk
    def save_to_disk(self, filename: str) -> None:
        """
        Saves graph to disk

        :Arguments:
            **filename** (:obj:`str`): Path to file. Usual file extension is *aeg*
        """
        mygraph = {}
        mygraph["description"] = self.description
        mygraph["num_links"] = self.num_links
        mygraph["num_nodes"] = self.num_nodes
        mygraph["network"] = self.network
        mygraph["graph"] = self.graph
        mygraph["all_nodes"] = self.all_nodes
        mygraph["nodes_to_indices"] = self.nodes_to_indices
        mygraph["num_nodes"] = self.num_nodes
        mygraph["fs"] = self.fs
        mygraph["cost"] = self.cost
        mygraph["cost_field"] = self.cost_field
        mygraph["skims"] = self.skims
        mygraph["skim_fields"] = self.skim_fields
        mygraph["block_centroid_flows"] = self.block_centroid_flows
        mygraph["centroids"] = self.centroids
        mygraph["graph_id"] = self._id
        mygraph["mode"] = self.mode

        with open(filename, "wb") as f:
            pickle.dump(mygraph, f)

    def load_from_disk(self, filename: str) -> None:
        """
        Loads graph from disk

        :Arguments:
            **filename** (:obj:`str`): Path to file
        """
        with open(filename, "rb") as f:
            mygraph = pickle.load(f)
            self.description = mygraph["description"]
            self.num_links = mygraph["num_links"]
            self.num_nodes = mygraph["num_nodes"]
            self.network = mygraph["network"]
            self.graph = mygraph["graph"]
            self.all_nodes = mygraph["all_nodes"]
            self.nodes_to_indices = mygraph["nodes_to_indices"]
            self.num_nodes = mygraph["num_nodes"]
            self.fs = mygraph["fs"]
            self.cost = mygraph["cost"]
            self.cost_field = mygraph["cost_field"]
            self.skims = mygraph["skims"]
            self.skim_fields = mygraph["skim_fields"]
            self.block_centroid_flows = mygraph["block_centroid_flows"]
            self.centroids = mygraph["centroids"]
            self._id = mygraph["graph_id"]
            self.mode = mygraph["mode"]
        self.__build_derived_properties()

    def __build_derived_properties(self):
        if self.centroids is None:
            return
        self.num_zones = self.centroids.shape[0] if self.centroids.shape else 0

    def available_skims(self) -> List[str]:
        """
        Returns graph fields that are available to be set as skims

        :Returns:
            **list** (:obj:`str`): Field names
        """
        return [x for x in self.graph.columns if x not in ["link_id", "a_node", "b_node", "direction", "id"]]

    # We check if all minimum fields are there
    def __network_error_checking__(self):
        # Checking field names
        has_fields = self.network.columns
        must_fields = ["link_id", "a_node", "b_node", "direction"]
        for field in must_fields:
            if field not in has_fields:
                raise ValueError(f"could not find field {field} in the network array")

        # Uniqueness of the id
        link_ids = self.network["link_id"].astype(int)
        if link_ids.shape[0] != np.unique(link_ids).shape[0]:
            raise ValueError('"link_id" field not unique')

            # Direction values
        if np.max(self.network["direction"]) > 1 or np.min(self.network["direction"]) < -1:
            raise ValueError('"direction" field not limited to (-1,0,1) values')

        if "id" not in self.network.columns:
            self.network = self.network.assign(id=np.nan)

    def __determine_types__(self, new_type, current_type):
        if new_type.isdigit():
            new_type = int(new_type)
        else:
            try:
                new_type = float(new_type)
            except ValueError as verr:
                self.logger.warning("Could not convert {} - {}".format(new_type, verr.__str__()))
        if isinstance(new_type, int):
            def_type = int
            if current_type == float:
                def_type = float
            elif current_type == str:
                def_type = str
        elif isinstance(new_type, float):
            def_type = float
            if current_type == str:
                def_type = str
        elif isinstance(new_type, str):
            def_type = str
        else:
            raise ValueError("WRONG TYPE OR NULL VALUE")
        return def_type

    def save_compressed_correspondence(self, path, mode_name, mode_id):
        """Save graph and nodes_to_indices to disk"""
        graph_path = join(path, f"correspondence_c{mode_name}_{mode_id}.feather")
        self.graph.to_feather(graph_path)
        node_path = join(path, f"nodes_to_indices_c{mode_name}_{mode_id}.feather")
        pd.DataFrame(self.nodes_to_indices, columns=["node_index"]).to_feather(node_path)

    def create_compressed_link_network_mapping(self):
        """
        Create two arrays providing a mapping of compressed id to link id.

        Uses sparse compression. Index ``idx`` by the by compressed id and compressed id + 1, the
        network IDs are then in the range ``idx[id]:idx[id + 1]``.

        .. code-block:: python

            >>> idx, data = graph.compressed_link_network_mapping
            >>> data[idx[id]:idx[id + 1]]  # ==> Slice of network ID's corresponding to the compressed ID

        Links not in the compressed graph are not contained within the ``data`` array.

        :Returns:
            **idx** (:obj:`np.array`): index array for ``data``
            **data** (:obj:`np.array`): array of link ids
        """

        return create_compressed_link_network_mapping(self)


class Graph(GraphBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TransitGraph(GraphBase):
    def __init__(self, config: dict = None, od_node_mapping: pd.DataFrame = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._config = config
        self.od_node_mapping = od_node_mapping
        self.mode = "t"



from sqlite3 import IntegrityError

from aequilibrae.project.field_editor import FieldEditor
from aequilibrae.project.network.link_type import LinkType
from aequilibrae.project.table_loader import TableLoader
from aequilibrae.utils.db_utils import commit_and_close
from aequilibrae.utils.spatialite_utils import connect_spatialite


class LinkTypes:
    """
    Access to the API resources to manipulate the link_types table in the network.

    .. code-block:: python

        >>> from aequilibrae import Project

        >>> p = Project.from_path("/tmp/test_project")

        >>> link_types = p.network.link_types

        # We can get a dictionary of link types in the model
        >>> all_link_types = link_types.all_types()

        # And do a bulk change and save it
        >>> for link_type_id, link_type_obj in all_link_types.items():
        ...     link_type_obj.beta = 1

        # We can save changes for all link types in one go
        >>> link_types.save()

        # or just get one link_type in specific
        >>> default_link_type = link_types.get('y')

        # or just get it by name
        >>> default_link_type = link_types.get_by_name('default')

        # We can change the description of the link types
        >>> default_link_type.description = 'My own new description'

        # Let's say we are using alpha to store lane capacity during the night as 90% of the standard
        >>> default_link_type.alpha = 0.9 * default_link_type.lane_capacity

        # To save this link types we can simply
        >>> default_link_type.save()

        # We can also create a completely new link_type and add to the model
        >>> new_type = link_types.new('a')
        >>> new_type.link_type = 'Arterial'  # Only ASCII letters and *_* allowed # other fields are not mandatory

        # We then save it to the database
        >>> new_type.save()

        # we can even keep editing and save it directly once we have added it to the project
        >>> new_type.lanes = 3
        >>> new_type.lane_capacity = 1100
        >>> new_type.save()

    """

    def __init__(self, net):
        self.__items = {}
        self.project = net.project
        self.logger = net.project.logger

        tl = TableLoader()
        with commit_and_close(connect_spatialite(self.project.path_to_file)) as conn:
            link_types_list = tl.load_table(conn, "link_types")
        existing_list = [lt["link_type_id"] for lt in link_types_list]

        self.__fields = list(tl.fields)
        for lt in link_types_list:
            if lt["link_type_id"] not in self.__items:
                self.__items[lt["link_type_id"]] = LinkType(lt, self.project)

        to_del = [key for key in self.__items.keys() if key not in existing_list]
        for key in to_del:
            del self.__items[key]

    def new(self, link_type_id: str) -> LinkType:
        if link_type_id in self.__items:
            raise ValueError(f"Link Type ID ({link_type_id}) already exists in the model. It must be unique.")

        tp = {key: None for key in self.__fields}
        tp["link_type_id"] = link_type_id
        lt = LinkType(tp, self.project)
        self.__items[link_type_id] = lt
        return lt

    def delete(self, link_type_id: str) -> None:
        """Removes the link_type with *link_type_id* from the project"""
        try:
            lt = self.__items[link_type_id]  # type: LinkType
            lt.delete()
            del self.__items[link_type_id]
        except IntegrityError as e:
            self.logger.error(f"Failed to remove link_type {link_type_id}. {e.args}")
            raise e
        self.logger.warning(f"Link type {link_type_id} was successfully removed from the project database")

    def get(self, link_type_id: str) -> LinkType:
        """Get a link_type from the network by its *link_type_id*"""
        if link_type_id not in self.__items:
            raise ValueError(f"Link type {link_type_id} does not exist in the model")
        return self.__items[link_type_id]

    def get_by_name(self, link_type: str) -> LinkType:
        """Get a link_type from the network by its *link_type* (i.e. name)"""
        for lt in self.__items.values():
            if lt.link_type.lower() == link_type.lower():
                return lt

    def fields(self) -> FieldEditor:
        """Returns a FieldEditor class instance to edit the Link_Types table fields and their metadata"""
        return FieldEditor(self.project.project_base_path, "link_types")

    def all_types(self) -> dict:
        """Returns a dictionary with all LinkType objects available in the model. link_type_id as key"""
        return self.__items

    def save(self):
        for lt in self.__items.values():  # type: LinkType
            lt.save()

    def __copy__(self):
        raise Exception("Link Types object cannot be copied")

    def __deepcopy__(self, memodict=None):
        raise Exception("Link Types object cannot be copied")

    def __del__(self):
        self.__items.clear()



import zipfile
import os
from os.path import dirname, join
from pathlib import Path
from typing import List

from aequilibrae.project import Project


def create_example(path: str, from_model="sioux_falls") -> Project:
    """Copies an example model to a new project project and returns the project handle

    :Arguments:
        **path** (:obj:`str`): Path where to create a new model. must be a non-existing folder/directory.

        **from_model** (:obj:`str`, *Optional*): Example to create from *sioux_falls*, *nauru* or *coquimbo*.
        Defaults to *sioux_falls*

    :Returns:
        **project** (:obj:`Project`): Aequilibrae Project handle (open)

    """
    if os.path.isdir(path):
        raise FileExistsError("Cannot overwrite an existing directory")

    if not os.path.isfile(join(dirname(__file__), f"../reference_files/{from_model}.zip")):
        raise FileExistsError("Example not found")

    os.makedirs(path, exist_ok=True)
    zipfile.ZipFile(join(dirname(__file__), f"../reference_files/{from_model}.zip")).extractall(path)
    proj = Project()
    proj.open(path)
    return proj


def list_examples() -> List[str]:
    pth = Path(__file__).parent.parent / "reference_files"
    return [str(x.stem) for x in pth.glob("*.zip")]



from copy import deepcopy
from os.path import join, realpath
from typing import Union, Dict

import shapely.wkb
from shapely.geometry import Point, Polygon, LineString, MultiLineString
from shapely.ops import unary_union

from aequilibrae.project.basic_table import BasicTable
from aequilibrae.project.project_creation import run_queries_from_sql_file
from aequilibrae.project.table_loader import TableLoader
from aequilibrae.project.zone import Zone
from aequilibrae.utils.db_utils import commit_and_close
from aequilibrae.utils.geo_index import GeoIndex
from aequilibrae.utils.spatialite_utils import connect_spatialite


class Zoning(BasicTable):
    """
    Access to the API resources to manipulate the zones table in the project

    .. code-block:: python

        >>> from aequilibrae import Project

        >>> project = Project.from_path("/tmp/test_project")


        >>> zoning = project.zoning

        >>> zone_downtown = zoning.get(1)
        >>> zone_downtown.population = 637
        >>> zone_downtown.employment = 10039
        >>> zone_downtown.save()

        # changing the value for an existing value/field
        >>> project.about.scenario_name = 'Just a better scenario name'
        >>> project.about.write_back()

        # We can also add one more field to the table
        >>> fields = zoning.fields
        >>> fields.add('parking_spots', 'Total licensed parking spots', 'INTEGER')
    """

    def __init__(self, network):
        super().__init__(network.project)
        self.__items: Dict[int, Zone] = {}
        self.network = network
        self.__table_type__ = "zones"
        self.__fields = []
        self.__geo_index = GeoIndex()
        if self.__has_zoning():
            self.__load()

    def new(self, zone_id: int) -> Zone:
        """Creates a new zone

        :Returns:
            **zone** (:obj:`Zone`): A new zone object populated only with zone_id (but not saved in the model yet)
        """

        if zone_id in self.__items:
            raise Exception(f"Zone ID {zone_id} already exists")

        data = {key: None for key in self.__fields}
        data["zone_id"] = zone_id

        self.project.logger.info(f"Zone with id {zone_id} was created")
        return self.__create_return_zone(data)

    def create_zoning_layer(self):
        """Creates the 'zones' table for project files that did not previously contain it"""

        if not self.__has_zoning():
            qry_file = join(realpath(__file__), "database_specification", "tables", "zones.sql")
            with commit_and_close(connect_spatialite(self.project.path_to_file)) as conn:
                run_queries_from_sql_file(conn, self.project.logger, qry_file)
            self.__load()
        else:
            self.project.warning("zones table already exists. Nothing was done", Warning)

    def coverage(self) -> Polygon:
        """Returns a single polygon for the entire zoning coverage

        :Returns:
            **model coverage** (:obj:`Polygon`): Shapely (Multi)polygon of the zoning system.
        """
        with commit_and_close(connect_spatialite(self.project.path_to_file)) as conn:
            dt = conn.execute('Select ST_asBinary("geometry") from zones;').fetchall()
        polygons = [shapely.wkb.loads(x[0]) for x in dt]
        return unary_union(polygons)

    def get(self, zone_id: str) -> Zone:
        """Get a zone from the model by its **zone_id**"""
        if zone_id not in self.__items:
            raise ValueError(f"Zone {zone_id} does not exist in the model")
        return self.__items[zone_id]

    def all_zones(self) -> dict:
        """Returns a dictionary with all Zone objects available in the model. *zone_id* as key"""
        return self.__items

    def save(self):
        for item in self.__items.values():
            item.save()

    def get_closest_zone(self, geometry: Union[Point, LineString, MultiLineString]) -> int:
        """Returns the zone in which the given geometry is located.

            If the geometry is not fully enclosed by any zone, the zone closest to
            the geometry is returned

        :Arguments:
            **geometry** (:obj:`Point` or :obj:`LineString`): A Shapely geometry object

        :Returns:
            **zone_id** (:obj:`int`): ID of the zone applicable to the point provided
        """

        nearest = self.__geo_index.nearest(geometry, 10)
        dists = {}
        for zone_id in nearest:
            geo = self.__items[zone_id].geometry
            if geo.contains(geometry):
                return zone_id
            dists[geo.distance(geometry)] = zone_id
        return dists[min(dists.keys())]

    def refresh_geo_index(self):
        self.__geo_index.reset()
        for zone_id, zone in self.__items.items():
            self.__geo_index.insert(feature_id=zone_id, geometry=zone.geometry)

    def __has_zoning(self):
        with commit_and_close(connect_spatialite(self.project.path_to_file)) as conn:
            dt = conn.execute("SELECT name FROM sqlite_master WHERE type='table';").fetchall()
        return any("zone" in x[0].lower() for x in dt)

    def __load(self):
        tl = TableLoader()
        with commit_and_close(connect_spatialite(self.project.path_to_file)) as conn:
            zones_list = tl.load_table(conn, "zones")
        self.__fields = deepcopy(tl.fields)

        existing_list = [zn["zone_id"] for zn in zones_list]
        if zones_list:
            self.__properties = list(zones_list[0].keys())
        for zn in zones_list:
            if zn["zone_id"] not in self.__items:
                self.__items[zn["zone_id"]] = Zone(zn, self)

        to_del = [key for key in self.__items.keys() if key not in existing_list]
        for key in to_del:
            del self.__items[key]
        self.refresh_geo_index()

    def _remove_zone(self, zone_id: int):
        del self.__items[zone_id]

    def __create_return_zone(self, data):
        zone = Zone(data, self)
        self.__items[zone.zone_id] = zone
        return zone



import string
import uuid
from os.path import join, dirname, realpath

from aequilibrae.project.project_creation import run_queries_from_sql_file
from aequilibrae.utils.db_utils import commit_and_close
from aequilibrae.utils.spatialite_utils import connect_spatialite


class About:
    """Provides an interface for querying and editing the **about** table of an AequilibraE project

    .. code-block:: python

        >>> from aequilibrae import Project

        >>> project = Project.from_path("/tmp/test_project")

        # Adding a new field and saving it
        >>> project.about.add_info_field('my_super_relevant_field')
        >>> project.about.my_super_relevant_field = 'super relevant information'
        >>> project.about.write_back()

        # changing the value for an existing value/field
        >>> project.about.scenario_name = 'Just a better scenario name'
        >>> project.about.write_back()

    """

    def __init__(self, project):
        self.__characteristics = []
        self.__original = {}
        self.__path_to_file = project.path_to_file
        self.logger = project.logger

        with commit_and_close(connect_spatialite(self.__path_to_file)) as conn:
            if self.__has_about(conn):
                self.__load(conn)

    def create(self):
        """Creates the 'about' table for project files that did not previously contain it"""

        with commit_and_close(connect_spatialite(self.__path_to_file)) as conn:
            if not self.__has_about(conn):
                qry_file = join(dirname(realpath(__file__)), "database_specification", "tables", "about.sql")
                run_queries_from_sql_file(conn, self.logger, qry_file)

            sql = "SELECT count(*) as num_records from about;"
            if conn.execute(sql).fetchone()[0] == 0:
                conn.execute(f"UPDATE 'about' set infovalue='{uuid.uuid4().hex}' where infoname='project_ID'")
                conn.execute("UPDATE 'about' set infovalue='right' where infoname='driving_side'")
                self.__load(conn)
            else:
                self.logger.warning("About table already exists. Nothing was done")

    def list_fields(self) -> list:
        """Returns a list of all characteristics the about table holds"""

        return list(self.__characteristics)

    def add_info_field(self, info_field: str) -> None:
        """Adds new information field to the model

        :Arguments:
            **info_field** (:obj:`str`): Name of the desired information field to be added. Has to be a valid
            Python VARIABLE name (i.e. letter as first character, no spaces and no special characters)

        .. code-block:: python

            >>> from aequilibrae import Project

            >>> p = Project.from_path("/tmp/test_project")
            >>> p.about.add_info_field('a_cool_field')
            >>> p.about.a_cool_field = 'super relevant information'
            >>> p.about.write_back()
        """
        allowed = string.ascii_lowercase + "_"
        has_forbidden = [x for x in info_field if x not in allowed]

        if has_forbidden:
            raise ValueError(f"{info_field} is not valid as a metadata field. Should be a lower case ascii letter or _")

        with commit_and_close(connect_spatialite(self.__path_to_file)) as conn:
            conn.execute("INSERT INTO 'about' (infoname) VALUES(?)", [info_field])
        self.__characteristics.append(info_field)
        self.__original[info_field] = None

    def write_back(self):
        """Saves the information parameters back to the project database

        .. code-block:: python

            >>> from aequilibrae import Project

            >>> p = Project.from_path("/tmp/test_project")
            >>> p.about.description = 'This is the example project. Do not use for forecast'
            >>> p.about.write_back()
        """
        with commit_and_close(connect_spatialite(self.__path_to_file)) as conn:
            for k in self.__characteristics:
                v = self.__dict__[k]
                if v != self.__original[k]:
                    conn.execute("UPDATE 'about' set infovalue = ? where infoname=?", [v, k])
                    self.logger.info(f"Updated {k} on About_Table to {v}")

    def __has_about(self, conn):
        sql = "SELECT name FROM sqlite_master WHERE type='table';"
        return any("about" in x[0] for x in conn.execute(sql).fetchall())

    def __load(self, conn):
        self.__characteristics = []
        sql = "select infoname, infovalue from 'about'"
        for x in conn.execute(sql).fetchall():
            self.__characteristics.append(x[0])
            self.__dict__[x[0]] = x[1]
            self.__original[x[0]] = x[1]



from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp.unordered_set cimport unordered_set
from libcpp.unordered_map cimport unordered_map
from libcpp cimport bool
from libcpp.memory cimport unique_ptr

cimport numpy as np  # Numpy *must* be cimport'd BEFORE pyarrow.lib, there's nothing quite like Cython.
cimport pyarrow as pa
cimport pyarrow.lib as libpa

# std::linear_congruential_engine is not available in the Cython libcpp.random shim. We'll import it ourselves
# from libcpp.random cimport minstd_rand
from libc.stdint cimport *

cdef extern from "<random>" namespace "std" nogil:
    cdef cppclass random_device:
        ctypedef uint_fast32_t result_type
        random_device() except +
        result_type operator()() except +

    cdef cppclass minstd_rand:
        ctypedef uint_fast32_t result_type
        minstd_rand() except +
        minstd_rand(result_type seed) except +
        result_type operator()() except +
        result_type min() except +
        result_type max() except +
        void discard(size_t z) except +
        void seed(result_type seed) except +

# std::shuffle is not available in the Cython libcpp.algorithm shim. We'll import it ourselves
# from libcpp.algorithm cimport shuffle
cdef extern from "<algorithm>" namespace "std" nogil:
    void shuffle[RandomIt, URBG](RandomIt first, RandomIt last, URBG&& g) except +

# std::make_pair is not available in the Cython libcpp.utilities shim. We'll import it ourselves based on the C++11 til C++14
# definition because C++14 makes this signature weird
# See https://github.com/cython/cython/issues/2706
cdef extern from "<utility>" namespace "std" nogil:
    pair[T, U] make_pair[T, U](T&& t, U&& u)

# To define our own hashing functions we have to write a little C++. The string is inlined directly into route_choice.cpp
# To make Cython aware of our hash types we also must declare them with the right signatures
#
# OrderedVectorPointerHasher: This hash function is for hashing the routes, thus it should be order *DEPENDENT*.
# Potential performance improvements may come from https://en.wikipedia.org/wiki/Universal_hashing#Hashing_vectors
#
# UnorderedSetPointerHasher: This hash function is for hashing the banned route sets, thus it should be order *INDEPENDENT*.
# Potential performance improvements may come from:
# Mark N. Wegman, J.Lawrence Carter,
# New hash functions and their use in authentication and set equality
# https://doi.org/10.1016/0022-0000(81)90033-7
#
# PointerDereferenceEqualTo: Because we are storing and hashing the pointers to objects to avoid unnecessary copies we must
# define our own comparator to resolve hash collisions. Without this equality operator the bare pointers are compared.
cdef extern from * nogil:
    """
    // Source: https://stackoverflow.com/a/72073933
    // License: CC BY-SA 4.0 Deed, https://creativecommons.org/licenses/by-sa/4.0/
    struct OrderedVectorPointerHasher {
        size_t operator()(const std::vector<long long> *V) const {
            size_t seed = V->size();
            long long x;
            for(auto &i : *V) {
                x = ((i >> 16) ^ i) * 0x45d9f3b;
                x = ((x >> 16) ^ x) * 0x45d9f3b;
                x = (x >> 16) ^ x;
                seed ^= x + 0x9e3779b9 + (seed << 6) + (seed >> 2);
            }
            return seed;
        }
    };

    // Source: https://stackoverflow.com/a/1537189
    // License: CC BY-SA 2.5 Deed, https://creativecommons.org/licenses/by-sa/2.5/
    struct UnorderedSetPointerHasher {
        size_t operator()(const std::unordered_set<long long> *S) const {
            size_t hash = 1;
            for(auto &i : *S) {
                hash *= 1779033703 + 2 * i;
            }
            return hash / 2;
        }
    };

    template<class T>
    struct PointerDereferenceEqualTo {
        bool operator()(const T& lhs, const T& rhs) const {
            return *lhs == *rhs;
        }
    };
    """
    cppclass OrderedVectorPointerHasher:
        size_t operator()(const vector[long long] *V) const

    cppclass UnorderedSetPointerHasher:
        size_t operator()(const unordered_set[long long] *S) const

    cppclass PointerDereferenceEqualTo[T]:
        bool operator()(const T& lhs, const T& rhs) const


# Pyarrow's Cython API does not provide all the functions available in the C++ API, some of them are really useful.
# Here we redeclare the classes with the functions we want, these are available in the current namespace, *not* libarrow
cdef extern from "arrow/builder.h" namespace "arrow" nogil:

    cdef cppclass CUInt32Builder" arrow::UInt32Builder"(libpa.CArrayBuilder):
        CUInt32Builder(libpa.CMemoryPool* pool)
        libpa.CStatus Append(const uint32_t value)
        libpa.CStatus AppendValues(const vector[uint32_t] &values)
        libpa.CStatus AppendValues(vector[uint32_t].const_reverse_iterator values_begin, vector[uint32_t].const_reverse_iterator values_end)
        libpa.CStatus AppendValues(const uint32_t *values, int64_t length, const uint8_t *valid_bytes = nullptr)

    cdef cppclass CDoubleBuilder" arrow::DoubleBuilder"(libpa.CArrayBuilder):
        CDoubleBuilder(libpa.CMemoryPool* pool)
        libpa.CStatus Append(const double value)
        libpa.CStatus AppendValues(const vector[double] &values)

    cdef cppclass CBooleanBuilder" arrow::BooleanBuilder"(libpa.CArrayBuilder):
        CBooleanBuilder(libpa.CMemoryPool* pool)
        libpa.CStatus Append(const bool value)
        libpa.CStatus AppendValues(const vector[bool] &values)


# For typing convenience, the types names are getting long
ctypedef unordered_set[vector[long long] *, OrderedVectorPointerHasher, PointerDereferenceEqualTo[vector[long long] *]] RouteSet_t
ctypedef unordered_set[unordered_set[long long] *, UnorderedSetPointerHasher, PointerDereferenceEqualTo[unordered_set[long long] *]] LinkSet_t
ctypedef vector[pair[unordered_set[long long] *, vector[long long] *]] RouteMap_t

ctypedef vector[unique_ptr[vector[long long]]] RouteVec_t

# A (known 2016) bug in the Cython compiler means it incorrectly parses the following type when used in a cdef
# https://github.com/cython/cython/issues/534
ctypedef vector[bool]* vector_bool_ptr



import warnings
import multiprocessing as mp

cimport cython
import numpy as np
from cython.parallel import parallel, prange
from libc.stdlib cimport malloc, free


def ipf_core(seed_matrix, target_productions, target_attractions, max_iterations=200, tolerance=0.001, cores = 0, warn=True):
    assert target_attractions.shape[0] == seed_matrix.shape[1]
    assert target_productions.shape[0] == seed_matrix.shape[0]

    cdef int max_iter = max_iterations
    cdef double toler = tolerance
    cdef int cpus = 0

    if cores < 0:
        cpus = max(1, mp.cpu_count() + cores)
    if cores == 0:
        cpus = mp.cpu_count()
    elif cores > 0:
        cpus = min(mp.cpu_count(), cores)

    mat_prod_tot = np.zeros_like(target_productions, np.float64)
    factor_prod = np.zeros_like(target_productions, np.float64)

    mat_attr_tot = np.zeros_like(target_attractions, np.float64)
    factor_attr = np.zeros_like(target_attractions, np.float64)

    cdef double [:] prod_tot = mat_prod_tot
    cdef double [:] prod_tgt = target_productions
    cdef double [:] prod_factor = factor_prod

    cdef double [:] attr_tot = mat_attr_tot
    cdef double [:] attr_tgt = target_attractions
    cdef double [:] attr_factor = factor_attr

    if seed_matrix.dtype == np.float32:
        iter, err =  _ipf_core32(seed_matrix, prod_tot, prod_tgt, prod_factor, attr_tot, attr_tgt, attr_factor, max_iter, toler, cpus)
    else:
        iter, err =  _ipf_core64(seed_matrix, prod_tot, prod_tgt, prod_factor, attr_tot, attr_tgt, attr_factor, max_iter, toler, cpus)

    if err > tolerance and warn:
        warnings.warn(f"Could not reach convergence in {iter} iterations: {err}")
    return iter, err

cdef _ipf_core64(seed_matrix, prod_tot, prod_tgt, prod_factor, attr_tot, attr_tgt, attr_factor, max_iter, toler, cpus):

    cdef double [:, :] flows = seed_matrix
    return _fratar(flows, prod_tot, prod_tgt, prod_factor, attr_tot, attr_tgt, attr_factor, max_iter, toler, cpus)


cdef _ipf_core32(seed_matrix, prod_tot, prod_tgt, prod_factor, attr_tot, attr_tgt, attr_factor, max_iter, toler, cpus):

    cdef float [:, :] flows_single = seed_matrix
    return _fratar_sing(flows_single, prod_tot, prod_tgt, prod_factor, attr_tot, attr_tgt, attr_factor, max_iter, toler, cpus)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef  _fratar(double[:, :] flows,
                   double[:] prod_tot,
                   double[:] prod_tgt,
                   double[:] prod_factor,
                   double[:] attr_tot,
                   double[:] attr_tgt,
                   double[:] attr_factor,
                   int max_iter,
                   double toler,
                   int cpus) noexcept:

    cdef double err = 1.0
    cdef int iter = 0
    cdef long long i, j
    cdef long long I = flows.shape[0]
    cdef long long J = flows.shape[1]

    # we zero everyone that needs to be zero to be able to skip them in the future
    for i in prange(I, nogil=True):
        for j in range(J):
            if prod_tgt[i] + attr_tgt[j] == 0:
                flows[i, j] = 0

    for iter in range(max_iter):
        _total_prods(flows, prod_tgt, prod_tot, cpus)
        err = _factors(prod_tgt, prod_tot, prod_factor, cpus)
        for i in prange(I, nogil=True, num_threads=cpus):
            if prod_tgt[i] > 0:
                for j in range(J):
                    flows[i, j] = flows[i, j] * prod_factor[i]


        _total_attra(flows, prod_tgt, attr_tot, cpus)
        err = _factors(attr_tgt, attr_tot, attr_factor, cpus)
        for i in prange(I, nogil=True, num_threads=cpus):
            if prod_tgt[i] > 0:
                for j in range(J):
                    flows[i, j] = flows[i, j] * attr_factor[j]

        # FUNCTION TO COMPUTE THE ERROR
        err = _calc_err(prod_factor, attr_factor)
        if (err - 1) < toler:
            break

    return iter, err - 1

@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void _total_attra(double[:, :] flows,
                        double[:] prod_tgt,
                        double[:] attr_tot,
                        int cpus) noexcept:

    cdef long long i, j, jk
    cdef double  *local_buf
    cdef long long I = flows.shape[0]
    cdef long long J = flows.shape[1]

    # Computes factors
    with nogil, parallel( num_threads=cpus):
        local_buf = <double *> malloc(sizeof(double) * J)
        for j in range(J):
            local_buf[j] = 0
            attr_tot[j] = 0

        for i in prange(I):
            if prod_tgt[i] == 0:
                continue
            for jk in range(J):
                # for jk in array_of_indices_of_non_zeros:
                local_buf[jk] += flows[i, jk]

        with gil:
            for j in range(J):
                attr_tot[j] += local_buf[j]

        free(local_buf)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void _total_prods(double[:, :] flows,
                        double[:] prod_tgt,
                        double[:] prod_tot,
                        int cpus) noexcept nogil:

    cdef long long i, j
    cdef long long I = flows.shape[0]
    cdef long long J = flows.shape[1]

    # Calculate the row totals (prods) from the flows
    for i in prange(I, num_threads=cpus):
        prod_tot[i] = 0
        if prod_tgt[i] == 0:
            continue
        for j in range(J):
            prod_tot[i] += flows[i, j]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef double _factors(double[:] target,
                      double[:] total,
                      double[:] factor,
                      int cpus) noexcept:

    cdef long long i, I = target.shape[0]
    cdef double err = 1.0

    # Computes factors
    with nogil, parallel(num_threads=cpus):
        for i in prange(I):
            factor[i] = 0
            if target[i] > 0:
                if total[i] == 0:
                    err = -1.0
                else:
                    factor[i] = target[i] / total[i]
    return err

@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef double _calc_err(double[:] p_factor,
                       double[:] a_factor) noexcept:

    cdef long long i, I = p_factor.shape[0]
    cdef long long j, J = a_factor.shape[0]
    cdef double err = 1.0

    # Production
    for i in range(I):
        if p_factor[i] > 0:
            err = err if err > 1 / p_factor[i] else 1 / p_factor[i]
        err = err if err > p_factor[i] else p_factor[i]

    # Attraction
    for j in range(J):
        if a_factor[j] > 0:
            err = err if err > 1 / a_factor[j] else 1 / a_factor[j]
        err = err if err > a_factor[j] else a_factor[j]
    return err


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef  _fratar_sing(float[:, :] flows,
                   double[:] prod_tot,
                   double[:] prod_tgt,
                   double[:] prod_factor,
                   double[:] attr_tot,
                   double[:] attr_tgt,
                   double[:] attr_factor,
                   int max_iter,
                   double toler,
                   int cpus) noexcept:

    cdef double err = 1.0
    cdef int iter = 0
    cdef long long i, j
    cdef long long I = flows.shape[0]
    cdef long long J = flows.shape[1]

    # we zero everyone that needs to be zero to be able to skip them in the future
    for i in prange(I, nogil=True):
        for j in range(J):
            if prod_tgt[i] + attr_tgt[j] == 0:
                flows[i, j] = 0

    for iter in range(max_iter):
        _total_prods_sing(flows, prod_tgt, prod_tot, cpus)
        err = _factors(prod_tgt, prod_tot, prod_factor, cpus)
        for i in prange(I, nogil=True, num_threads=cpus):
            if prod_tgt[i] > 0:
                for j in range(J):
                    flows[i, j] = flows[i, j] * prod_factor[i]


        _total_attra_sing(flows, prod_tgt, attr_tot, cpus)
        err = _factors(attr_tgt, attr_tot, attr_factor, cpus)
        for i in prange(I, nogil=True, num_threads=cpus):
            if prod_tgt[i] > 0:
                for j in range(J):
                    flows[i, j] = flows[i, j] * attr_factor[j]

        # FUNCTION TO COMPUTE THE ERROR
        err = _calc_err(prod_factor, attr_factor)
        if (err - 1) < toler:
            break

    return iter, err - 1


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void _total_attra_sing(float[:, :] flows,
                        double[:] prod_tgt,
                        double[:] attr_tot,
                        int cpus) noexcept:

    cdef long long i, j, jk
    cdef double  *local_buf
    cdef long long I = flows.shape[0]
    cdef long long J = flows.shape[1]

    # Computes factors
    with nogil, parallel( num_threads=cpus):
        local_buf = <double *> malloc(sizeof(double) * J)
        for j in range(J):
            local_buf[j] = 0
            attr_tot[j] = 0

        for i in prange(I):
            if prod_tgt[i] == 0:
                continue
            for jk in range(J):
                # for jk in array_of_indices_of_non_zeros:
                local_buf[jk] += flows[i, jk]

        with gil:
            for j in range(J):
                attr_tot[j] += local_buf[j]

        free(local_buf)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void _total_prods_sing(float[:, :] flows,
                        double[:] prod_tgt,
                        double[:] prod_tot,
                        int cpus) noexcept nogil:

    cdef long long i, j
    cdef long long I = flows.shape[0]
    cdef long long J = flows.shape[1]

    # Calculate the row totals (prods) from the flows
    for i in prange(I, num_threads=cpus):
        prod_tot[i] = 0
        if prod_tgt[i] == 0:
            continue
        for j in range(J):
            prod_tot[i] += flows[i, j]


from os import PathLike

import pandas as pd
import shapely.wkb

from aequilibrae.utils.db_utils import commit_and_close
from aequilibrae.utils.spatialite_utils import connect_spatialite


class DataLoader:
    def __init__(self, path_to_file: PathLike, table_name: str):
        self.__pth_file = path_to_file
        self.table_name = table_name

    def load_table(self) -> pd.DataFrame:
        with commit_and_close(connect_spatialite(self.__pth_file)) as conn:
            fields, _, geo_field = self.__find_table_fields()
            fields = [f'"{x}"' for x in fields]
            if geo_field is not None:
                fields.append('ST_AsBinary("geometry") geometry')
            keys = ",".join(fields)
            df = pd.read_sql_query(f"select {keys} from '{self.table_name}'", conn)
        if geo_field is not None:
            df.geometry = df.geometry.apply(shapely.wkb.loads)
        return df

    def __find_table_fields(self):
        with commit_and_close(connect_spatialite(self.__pth_file)) as conn:
            structure = conn.execute(f"pragma table_info({self.table_name})").fetchall()
        geotypes = ["LINESTRING", "POINT", "POLYGON", "MULTIPOLYGON"]
        fields = [x[1].lower() for x in structure]
        geotype = geo_field = None
        for x in structure:
            if x[2].upper() in geotypes:
                geotype = x[2]
                geo_field = x[1]
                break
        if geo_field is not None:
            fields = [x for x in fields if x != geo_field.lower()]

        return fields, geotype, geo_field



import itertools
import warnings
import logging
import pathlib
import socket
import sqlite3
from datetime import datetime
from typing import List, Optional, Tuple, Union, Dict
from uuid import uuid4
from functools import cached_property

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.dataset
import scipy
from aequilibrae.context import get_active_project
from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.paths.graph import Graph, _get_graph_to_network_mapping
from aequilibrae.paths.cython.route_choice_set import RouteChoiceSet
from aequilibrae.paths.cython.coo_demand import GeneralisedCOODemand


class RouteChoice:
    all_algorithms = ["bfsle", "lp", "link-penalisation", "link-penalization"]

    default_parameters = {
        "generic": {
            "seed": 0,
            "max_routes": 0,
            "max_depth": 0,
            "max_misses": 100,
            "penalty": 1.01,
            "cutoff_prob": 0.0,
            "beta": 1.0,
            "store_results": True,
        },
        "link-penalisation": {},
        "bfsle": {"penalty": 1.0},
    }

    demand_index_names = ["origin id", "destination id"]

    def __init__(self, graph: Graph, project=None):
        self.parameters = self.default_parameters.copy()
        self.procedure_id = None
        self.procedure_date = None

        proj = project or get_active_project(must_exist=False)
        self.project = proj

        self.logger = proj.logger if proj else logging.getLogger("aequilibrae")

        self.cores: int = 0
        self.graph = graph
        self.demand = self.__init_demand()

        self.schema = RouteChoiceSet.schema
        self.psl_schema = RouteChoiceSet.psl_schema

        self.sl_compact_link_loads: Optional[Dict[str, np.array]] = None
        self.sl_link_loads: Optional[Dict[str, np.array]] = None

        self.results: Optional[pa.Table] = None
        self.where: Optional[pathlib.Path] = None
        self.save_path_files: bool = False

        self._config = {}
        self._selected_links = {}

    @cached_property
    def __rc(self) -> RouteChoiceSet:
        return RouteChoiceSet(self.graph)

    def __init_demand(self):
        d = GeneralisedCOODemand(
            *self.demand_index_names, self.graph.nodes_to_indices, shape=(self.graph.num_zones, self.graph.num_zones)
        )
        return d

    def set_choice_set_generation(self, /, algorithm: str, **kwargs) -> None:
        """Chooses the assignment algorithm and set parameters.
        Options for algorithm are, 'bfsle' for breadth first search with link removal, or 'link-penalisation'/'link-penalization'.

        BFSLE implementation based on "Route choice sets for very high-resolution data" by Nadine Rieser-Schüssler,
        Michael Balmer & Kay W. Axhausen (2013).
        https://doi.org/10.1080/18128602.2012.671383

        'lp' is also accepted as an alternative to 'link-penalisation'

        Setting the parameters for the route choice:

        `seed` is a BFSLE specific parameters.

        Setting `max_depth` or `max_misses`, while not required, is strongly recommended to prevent runaway algorithms.
        `max_misses` is the maximum amount of duplicate routes found per OD pair. If it is exceeded then the route set
        if returned with fewer than `max_routes`. It has a default value of `100`.

        - When using BFSLE `max_depth` corresponds to the maximum height of the graph of graphs. It's value is
            largely dependent on the size of the paths within the network. For very small networks a value of 10
            is a recommended starting point. For large networks a good starting value is 5. Increase the value
            until the number of desired routes is being consistently returned. If it is exceeded then the route set
            if returned with fewer than `max_routes`.

        - When using LP, `max_depth` corresponds to the maximum number of iterations performed. While not enforced,
            it should be higher than `max_routes`. It's value is dependent on the magnitude of the cost field,
            specifically it's related to the log base `penalty` of the ratio of costs between two alternative routes.
            If it is exceeded then the route set if returned with fewer than `max_routes`.

        Additionally BFSLE has the option to incorporate link penalisation. Every link in all routes found at a depth
        are penalised with the `penalty` factor for the next depth. So at a depth of 0 no links are penalised nor
        removed. At depth 1, all links found at depth 0 are penalised, then the links marked for removal are removed.
        All links in the routes found at depth 1 are then penalised for the next depth. The penalisation compounds.
        Pass set `penalty=1.0` to disable.

        When performing an assignment, `cutoff_prob` can be provided to exclude routes from the path-sized logit model.
        The `cutoff_prob` is used to compute an inverse binary logit and obtain a max difference in utilities. If a
        paths total cost is greater than the minimum cost path in the route set plus the max difference, the route is
        excluded from the PSL calculations. The route is still returned, but with a probability of 0.0.

        The `cutoff_prob` should be in the range [0, 1]. It is then rescaled internally to [0.5, 1] as probabilities
        below 0.5 produce negative differences in utilities because the choice is between two routes only, one of
        which is the shortest path. A higher `cutoff_prob` includes less routes. A value of `1.0` will only include
        the minimum cost route. A value of `0.0` includes all routes.

        :Arguments:
            **algorithm** (:obj:`str`): Algorithm to be used
            **kwargs** (:obj:`dict`): Dictionary with all parameters for the algorithm
        """
        algo_dict = {i: i for i in self.all_algorithms}
        algo_dict["lp"] = "link-penalisation"
        algo_dict["link-penalization"] = "link-penalisation"
        algo = algo_dict.get(algorithm.lower())

        if algo is None:
            raise AttributeError(f"Assignment algorithm not available. Choose from: {','.join(self.all_algorithms)}")

        defaults = self.default_parameters["generic"] | self.default_parameters[algo]
        for key in kwargs.keys():
            if key not in defaults:
                raise ValueError(f"Invalid parameter `{key}` provided for algorithm `{algo}`")

        self.algorithm = algo
        self._config["Algorithm"] = algo

        self.parameters = defaults | kwargs

    def set_cores(self, cores: int) -> None:
        """Allows one to set the number of cores to be used

            Inherited from :obj:`AssignmentResultsBase`

        :Arguments:
            **cores** (:obj:`int`): Number of CPU cores to use
        """
        self.cores = cores

    def set_save_path_files(self, save_it: bool) -> None:
        """turn path saving on or off.

        :Arguments:
            **save_it** (:obj:`bool`): Boolean to indicate whether paths should be saved
        """
        self.save_path_files = save_it
        raise NotImplementedError()

    def set_save_routes(self, where: Optional[str] = None) -> None:
        """
        Set save path for route choice results. Provide ``None`` to disable.

        **warning** enabling route saving will disable in memory results. Viewing the results will read the results
        from disk first.

        :Arguments:
            **save_it** (:obj:`bool`): Boolean to indicate whether routes should be saved
        """

        if where is not None:
            where = pathlib.Path(where)
            if not where.exists():
                raise ValueError(f"Path does not exist `{where}`")
        self.where = where

    def add_demand(self, demand, fill: float = 0.0):
        """
        Add demand DataFrame or matrix for the assignment.

        :Arguments:
            **demand** (:obj:`Union[pd.DataFrame, AequilibraeMatrix]`): Demand to add to assignment. If the supplied
              demand is a DataFrame, it should have a 2-level MultiIndex of Origin and Destination node IDs. If an
              AequilibraE matrix is supplied node IDs will be inferred from the index. Demand values should be either
              float32s or float64s.

            **fill** (:obj:`float`): Value to fill any NaNs with.
        """
        if isinstance(demand, pd.DataFrame):
            self.demand.add_df(demand, fill=fill)
        elif isinstance(demand, AequilibraeMatrix):
            self.demand.add_matrix(demand, fill=fill)
        else:
            raise TypeError(f"unknown argument type '{(type(demand).__name__)}'")

    def prepare(self, nodes: Union[List[int], List[Tuple[int, int]], None] = None) -> None:
        """
        Prepare OD pairs for batch computation.

        :Arguments:
            **nodes** (:obj:`Union[list[int], list[tuple[int, int]]]`): List of node IDs to operate on. If a 1D list is
                provided, OD pairs are taken to be all pair permutations of the list. If a list of pairs is provided
                OD pairs are taken as is. All node IDs must be present in the compressed graph. To make a node ID
                always appear in the compressed graph add it as a centroid. Duplicates will be dropped on execution.
                If *None* is provided, all OD pairs with non-zero flows will be used.
        """
        if nodes is not None and not self.demand.no_demand():
            raise ValueError("provide either `nodes` or set a `demand` matrix, not both")
        elif nodes is None:
            return
        elif len(nodes) == 0:
            raise ValueError("`nodes` list-like empty.")

        self.demand = self.__init_demand()
        df = pd.DataFrame()
        if all(
            isinstance(pair, tuple)
            and len(pair) == 2
            and isinstance(pair[0], (int, np.integer))
            and isinstance(pair[1], (int, np.integer))
            for pair in nodes
        ):
            df.index = pd.MultiIndex.from_tuples(nodes, name=self.demand_index_names)
        elif len(nodes) > 1 and all(isinstance(x, (int, np.unsignedinteger)) for x in nodes):
            df.index = pd.MultiIndex.from_tuples(itertools.permutations(nodes, r=2), names=self.demand_index_names)
        else:
            raise ValueError(f"{type(nodes)} or {type(nodes[0])} for not valid types for the `prepare` method")

        self.demand.add_df(df)

    def execute_single(self, origin: int, destination: int, demand: float = 0.0) -> List[Tuple[int]]:
        """
        Generate route choice sets between `origin` and `destination`, potentially performing an assignment.

        Does not require preparation.

        Node IDs must be present in the compressed graph. To make a node ID always appear in the compressed
        graph add it as a centroid.

        :Arguments:
            **origin** (:obj:`int`): Origin node ID.
            **destination** (:obj:`int`): Destination node ID.
            **demand** (:obj:`float`): If provided an assignment will be performed with this demand.

        :Returns:
            ***route set** (:obj:`List[Tuple[int]]`): A list of routes as tuples of link IDs.
        """
        self.procedure_id = uuid4().hex
        self.procedure_date = str(datetime.today())

        self.results = None
        return self.__rc.run(
            origin,
            destination,
            self.demand.shape,
            demand=demand,
            bfsle=self.algorithm == "bfsle",
            path_size_logit=bool(demand),
            cores=self.cores,
            where=str(self.where) if self.where is not None else None,
            **self.parameters,
        )

    def execute(self, perform_assignment: bool = True) -> None:
        """
        Generate route choice sets between the previously supplied nodes, potentially performing an assignment.

        To access results see `RouteChoice.get_results()`.

        :Arguments:
            **perform_assignment** (:obj:`bool`): Whether or not to perform an assignment. Default `False`.
        """
        if self.demand.df.index.empty:
            raise ValueError(
                "to perform batch route choice generation you must first prepare with the selected nodes. See `RouteChoice.prepare()`"
            )

        self.procedure_date = str(datetime.today())

        self.results = None
        self.__rc.batched(
            self.demand,
            self._selected_links,
            bfsle=self.algorithm == "bfsle",
            path_size_logit=perform_assignment,
            cores=self.cores,
            where=str(self.where) if self.where is not None else None,
            **self.parameters,
        )

    def info(self) -> dict:
        """Returns information for the transit assignment procedure

        Dictionary contains keys  'Algorithm', 'Matrix totals', 'Computer name', 'Procedure ID', 'Parameters', and
        'Select links'.

        The classes key is also a dictionary with all the user classes per transit class and their respective
        matrix totals

        :Returns:
            **info** (:obj:`dict`): Dictionary with summary information
        """

        matrix_totals = self.demand.df.sum().to_dict()

        info = {
            "Algorithm": self.algorithm,
            "Matrix totals": matrix_totals,
            "Computer name": socket.gethostname(),
            "Procedure ID": self.procedure_id,
            "Parameters": self.parameters,
            "Select links": self._selected_links,
        }
        return info

    def log_specification(self):
        self.logger.info("Route Choice specification")
        self.logger.info(self._config)

    def get_results(self) -> Union[pa.Table, pa.dataset.Dataset]:
        """Returns the results of the route choice procedure

        Returns a table of OD pairs to lists of link IDs for each OD pair provided (as columns).
        Represents paths from ``origin`` to ``destination``.

        If `save_routes` was specified then a Pyarrow dataset is returned. The caller is responsible for reading this dataset.

        :Returns:
            **results** (:obj:`pa.Table`): Table with the results of the route choice procedure
        """
        if self.results is None:
            try:
                self.results = self.__rc.get_results()
            except RuntimeError as err:
                if self.where is None:
                    raise ValueError("Route choice results not computed and read/save path not specified") from err
                self.results = pa.dataset.dataset(
                    self.where, format="parquet", partitioning=pa.dataset.HivePartitioning(self.schema)
                )

        return self.results

    def get_load_results(self) -> pd.DataFrame:
        """
        Translates the link loading results from the graph format into the network format.

        :Returns:
            **dataset** (:obj:`Union[Tuple[pd.DataFrame, pd.DataFrame], pd.DataFrame]`):
                A tuple of link loading results as DataFrames.
                Columns are the matrix name concatenated direction.
        """

        if self.demand.no_demand():
            raise ValueError("No demand was provided. To perform link loading add a demand matrix or data frame")

        ll = self.__rc.get_link_loading(cores=self.cores)

        # Create a data store with a row for each uncompressed link
        m = _get_graph_to_network_mapping(self.graph.graph.link_id.values, self.graph.graph.direction.values)
        lids = np.unique(self.graph.graph.link_id.values)
        df = self.__link_loads_to_df(m, lids, ll)

        return df

    def __link_loads_to_df(self, mapping, lids, link_loads):
        df = pd.DataFrame(
            {"link_id": lids} | {k + dir: np.zeros(lids.shape) for k in link_loads.keys() for dir in ["_ab", "_ba"]}
        )
        added_dfs = []
        for k, v in link_loads.items():
            # Directional Flows
            df.iloc[mapping.network_ab_idx, df.columns.get_loc(k + "_ab")] = np.nan_to_num(v[mapping.graph_ab_idx])
            df.iloc[mapping.network_ba_idx, df.columns.get_loc(k + "_ba")] = np.nan_to_num(v[mapping.graph_ba_idx])

            # Tot Flow
            added_dfs.append(pd.DataFrame({f"{k}_tot": df[k + "_ab"] + df[k + "_ba"]}))

        df = pd.concat([df] + added_dfs, axis=1)
        cols = df.columns.tolist()
        cols.remove("link_id")
        cols.sort()
        # Insert the certain value at the beginning
        cols.insert(0, "link_id")
        return df[cols]

    def set_select_links(self, links: Dict[str, List[Union[Tuple[int, int], List[Tuple[int, int]]]]]):
        """
        Set the selected links. Checks if the links and directions are valid. Supports OR and AND sets of links.

        Dictionary values should be a list of either (link_id, dir) or a list of (link_id, dir).

        The elements of the first list represent the AND sets, together they are OR'ed. If any of these sets is
        satisfied the link are loaded as appropriate. The AND sets are comprised of either a single (link_id, dir) tuple
        or a list of (link_id, dir). The single tuple represents an AND set with a single element. All links and
        directions in an AND set must appear in any order within a route for it to be considered satisfied.
        Supply `links=None` to disable select link analysis.

        :Arguments:
            **links** (:obj:`Union[None, Dict[str, List[Union[Tuple[int, int], List[Tuple[int, int]]]]]]`):
                Name of link set and Link IDs and directions to be used in select link analysis.
        """
        self._selected_links = {}

        if links is None:
            del self._config["select_links"]
            return

        max_id = self.graph.compact_graph.id.max() + 1

        for name, link_set in links.items():
            if len(name.split(" ")) != 1:
                warnings.warn("Input string name has a space in it. Replacing with _")
                name = str.join("_", name.split(" "))

            normalised_link_set = []
            for link_ids in link_set:
                if isinstance(link_ids, tuple) and len(link_ids) == 2 and link_ids[1] == 0:
                    warnings.warn(
                        f"Adding both directions of a link ({link_ids[0]}) to a single AND set is likely "
                        f"unintentional. Replacing with {(link_ids[0], -1)} OR {(link_ids[0], 1)}"
                    )
                    normalised_link_set.append((link_ids[0], -1))
                    normalised_link_set.append((link_ids[0], 1))
                else:
                    normalised_link_set.append(link_ids)

            or_set = set()
            for link_ids in normalised_link_set:
                # For compatibility, a tuple of length 2 is passed, we assume that's a single (link, dir) pair
                if isinstance(link_ids, tuple) and len(link_ids) == 2:
                    link_ids = (link_ids,)

                and_set = set()
                for link, dir in link_ids:
                    if dir == 0:
                        query = (self.graph.graph["link_id"] == link) & (
                            (self.graph.graph["direction"] == -1) | (self.graph.graph["direction"] == 1)
                        )
                    else:
                        query = (self.graph.graph["link_id"] == link) & (self.graph.graph["direction"] == dir)

                    if not query.any():
                        raise ValueError(f"link_id or direction {(link, dir)} is not present within graph.")

                    for comp_id in self.graph.graph[query]["__compressed_id__"].values:
                        # Check for duplicate compressed link ids in the current link set
                        if comp_id == max_id:
                            raise ValueError(
                                f"link ID {link} and direction {dir} is not present in compressed graph. "
                                "It may have been removed during dead-end removal."
                            )
                        elif comp_id in and_set:
                            warnings.warn(
                                "Two input links map to the same compressed link in the network"
                                f", removing superfluous link {link} and direction {dir} with compressed id {comp_id}"
                            )
                        else:
                            and_set.add(comp_id)

                or_set.add(frozenset(and_set))
            self._selected_links[name] = frozenset(or_set)
        self._config["select_links"] = str(links)

    def get_select_link_loading_results(self) -> pd.DataFrame:
        """
        Get the select link loading results.

        :Returns:
            **dataset** (:obj:`Tuple[pd.DataFrame, pd.DataFrame]`): Select link loading results as DataFrames.
                Columns are the matrix name concatenated with the select link set and direction.
        """

        if self.demand.no_demand():
            raise ValueError("No demand was provided. To perform link loading add a demand matrix or data frame")

        sl_link_loads = {}
        for sl_name, sl_res in self.__rc.get_sl_link_loading().items():
            for demand_name, res in sl_res.items():
                sl_link_loads[sl_name + "_" + demand_name] = res

        # Create a data store with a row for each uncompressed link
        m = _get_graph_to_network_mapping(self.graph.graph.link_id.values, self.graph.graph.direction.values)
        lids = np.unique(self.graph.graph.link_id.values)
        df = self.__link_loads_to_df(m, lids, sl_link_loads)

        return df

    def get_select_link_od_matrix_results(self) -> Dict[str, Dict[str, scipy.sparse.coo_matrix]]:
        """
        Get the select link OD matrix results as a sparse matrix.

        :Returns:
            **select link OD matrix results** (:obj:`Dict[str, Dict[str, scipy.sparse.coo_matrix]]`): Returns a dict of
                select link set names to a dict of demand column names to a sparse OD matrix
        """

        if self.demand.no_demand():
            raise ValueError("No demand was provided. To perform link loading add a demand matrix or data frame")

        return self.__rc.get_sl_od_matrices()

    def __save_dataframe(self, df, method_name: str, description: str, table_name: str, report: dict, project) -> None:
        self.procedure_id = uuid4().hex
        data = [
            table_name,
            "select link",
            self.procedure_id,
            str(report),
            self.procedure_date,
            description,
        ]

        # sqlite3 context managers only commit, they don't close, oh well
        conn = sqlite3.connect(pathlib.Path(project.project_base_path) / "results_database.sqlite")
        with conn:
            df.to_sql(table_name, conn, index=False)
        conn.close()

        conn = project.connect()
        with conn:
            conn.execute(
                """Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                                description) Values(?,?,?,?,?,?)""",
                data,
            )
        conn.close()

    def save_link_flows(self, table_name: str, project=None) -> None:
        """
        Saves the link link flows for all classes into the results database.

        :Arguments:
            **table_name** (:obj:`str`): Name of the table being inserted to.
            **project** (:obj:`Project`, `Optional`): Project we want to save the results to.
            Defaults to the active project
        """
        if not project:
            project = self.project or get_active_project()

        df = self.get_load_results()
        info = self.info()
        self.__save_dataframe(
            df,
            "Link loading",
            "Uncompressed link loading results",
            table_name + "_uncompressed",
            info,
            project=project,
        )

    def save_select_link_flows(self, table_name: str, project=None) -> None:
        """
        Saves the select link link flows for all classes into the results database. Additionally, it exports
        the OD matrices into OMX format.

        :Arguments:
            **table_name** (:obj:`str`): Name of the table being inserted to and the name of the
            OpenMatrix file used for OD matrices.
            **project** (:obj:`Project`, `Optional`): Project we want to save the results to.
            Defaults to the active project
        """
        if not project:
            project = self.project or get_active_project()

        u = self.get_select_link_loading_results()
        info = self.info()
        self.__save_dataframe(
            u,
            "Select link analysis",
            "Uncompressed select link analysis results",
            table_name + "_uncompressed",
            info,
            project=project,
        )

        for sl_name, v in self.get_select_link_od_matrix_results().items():
            for demand_name, mat in v.items():
                mat.to_disk(
                    (pathlib.Path(project.project_base_path) / "matrices" / table_name).with_suffix(".omx"),
                    sl_name + "_" + demand_name,
                )



import time
from datetime import timedelta, datetime
from typing import Union, Any, Optional

day_sequence = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]


def to_seconds(value: Optional[str]) -> Union[int, Any]:
    if value is None or not isinstance(value, str):
        return value

    day = 0.0
    if "day" in value:
        day = float(value[: value.find(" d")]) * 24 * 3600
        value = value[value.find(", ") + 2 :]

    # convert the string to second of day
    split = value.split(":")
    if len(split) != 3:
        raise ValueError(f"Time {value} does not have an appropriate format")

    hours, minutes, seconds = map(int, split)
    return hours * 3600 + minutes * 60 + seconds + day


def to_time_string(value: Union[int, None]) -> Union[str, None]:
    # it is ok to pass None
    if value is None:
        return value

    if isinstance(value, str):
        if value.isdigit():
            value = int(value)

    if not isinstance(value, int):
        raise TypeError(f"Time {value} is not integer, but it should")

    return str(timedelta(seconds=value))


def one_day_before(date_object):
    start_seconds = date_object.timestamp() + 12 * 60 * 60
    return time.strftime("%Y-%m-%d", time.localtime(start_seconds - 24 * 60 * 60))


def create_days_between(range_start_date, range_end_date):
    numdays = (range_end_date - range_start_date).days + 1
    return [(range_start_date + timedelta(days=x)).strftime("%Y-%m-%d") for x in range(numdays)]


def day_of_week(date: str):
    return datetime.fromisoformat(date).weekday()


def format_date(date: str) -> str:
    return "-".join([date[:4], date[4:6], date[6:]])



import multiprocessing as mp
import sys
import threading
from datetime import datetime
from multiprocessing.dummy import Pool as ThreadPool
from uuid import uuid4

from aequilibrae import global_logger
from aequilibrae.context import get_active_project
from aequilibrae.paths.multi_threaded_skimming import MultiThreadedNetworkSkimming
from aequilibrae.paths.results.skim_results import SkimResults

try:
    from aequilibrae.paths.AoN import skimming_single_origin
except ImportError as ie:
    global_logger.warning(f"Could not import procedures from the binary. {ie.args}")

from aequilibrae.utils.signal import SIGNAL

sys.dont_write_bytecode = True


class NetworkSkimming:
    """

    .. code-block:: python

        >>> from aequilibrae import Project
        >>> from aequilibrae.paths.network_skimming import NetworkSkimming

        >>> project = Project.from_path("/tmp/test_project")

        >>> network = project.network
        >>> network.build_graphs()

        >>> graph = network.graphs['c']
        >>> graph.set_graph(cost_field="distance")
        >>> graph.set_skimming("distance")

        >>> skm = NetworkSkimming(graph)
        >>> skm.execute()

        # The skim report (if any error generated) is available here
        >>> skm.report
        []

        # To access the skim matrix directly from its temporary file
        >>> matrix = skm.results.skims

        # Or you can save the results to disk
        >>> skm.save_to_project('/tmp/skimming result.omx')

        # Or specify the AequilibraE's matrix file format
        >>> skm.save_to_project('skimming result', 'aem')

        >>> project.close()
    """

    skimming = SIGNAL(object)

    def __init__(self, graph, origins=None, project=None):
        self.project = project
        self.origins = origins
        self.graph = graph
        self.cores = mp.cpu_count()
        self.results = SkimResults()
        self.aux_res = MultiThreadedNetworkSkimming()
        self.report = []
        self.procedure_id = ""
        self.procedure_date = ""
        self.cumulative = 0

    def doWork(self):
        self.execute()

    def execute(self):
        """Runs the skimming process as specified in the graph"""
        self.skimming.emit(["zones finalized", 0])
        self.results.cores = self.cores
        self.results.prepare(self.graph)
        self.aux_res = MultiThreadedNetworkSkimming()
        self.aux_res.prepare(self.graph, self.results)

        pool = ThreadPool(self.results.cores)
        all_threads = {"count": 0}
        for orig in list(self.graph.centroids):
            i = int(self.graph.nodes_to_indices[orig])
            if i >= self.graph.nodes_to_indices.shape[0]:
                self.report.append(f"Centroid {orig} is beyond the domain of the graph")
            elif self.graph.fs[int(i)] == self.graph.fs[int(i) + 1]:
                self.report.append(f"Centroid {orig} does not exist in the graph")
            else:
                pool.apply_async(self.__func_skim_thread, args=(orig, all_threads))
        pool.close()
        pool.join()
        self.aux_res = None
        self.procedure_id = uuid4().hex
        self.procedure_date = str(datetime.today())

        self.skimming.emit(["text skimming", "Saving Outputs"])
        self.skimming.emit(["finished_threaded_procedure", None])

    def set_cores(self, cores: int) -> None:
        """
        Sets number of cores (threads) to be used in computation

        Value of zero sets number of threads to all available in the system, while negative values indicate the number
        of threads to be left out of the computational effort.

        Resulting number of cores will be adjusted to a minimum of zero or the maximum available in the system if the
        inputs result in values outside those limits

        :Arguments:
            **cores** (:obj:`int`): Number of cores to be used in computation
        """

        if isinstance(cores, int):
            if cores < 0:
                self.cores = max(1, mp.cpu_count() + cores)
            if cores == 0:
                self.cores = mp.cpu_count()
            elif cores > 0:
                cores = min(mp.cpu_count(), cores)
                if self.cores != cores:
                    self.cores = cores
        else:
            raise ValueError("Number of cores needs to be an integer")

    def save_to_project(self, name: str, format="omx", project=None) -> None:
        """Saves skim results to the project folder and creates record in the database

        :Arguments:
            **name** (:obj:`str`): Name of the matrix. Same value for matrix record name and file (plus extension)

            **format** (:obj:`str`, *Optional*): File format ('aem' or 'omx'). Default is ``'omx'``

            **project** (:obj:`Project`, *Optional*): Project we want to save the results to.
            Defaults to the active project
        """

        file_name = f"{name}.{format.lower()}"
        if not project:
            project = self.project or get_active_project()
        mats = project.matrices
        record = mats.new_record(name, file_name, self.results.skims)
        record.procedure_id = self.procedure_id
        record.timestamp = self.procedure_date
        record.procedure = "Network skimming"
        record.save()

    def __func_skim_thread(self, origin, all_threads):
        if threading.get_ident() in all_threads:
            th = all_threads[threading.get_ident()]
        else:
            all_threads[threading.get_ident()] = all_threads["count"]
            th = all_threads["count"]
            all_threads["count"] += 1
        x = skimming_single_origin(origin, self.graph, self.results, self.aux_res, th)
        self.cumulative += 1
        if x != origin:
            self.report.append(x)

        self.skimming.emit(["zones finalized", self.cumulative])
        txt = str(self.cumulative) + " / " + str(self.matrix.zones)
        self.skimming.emit(["text skimming", txt])



"""Convenience functions for working with geospatial data.
"""

import numpy as np


def haversine(lon1, lat1, lon2, lat2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)
    """
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(dlat / 2.0) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2.0) ** 2

    c = 2.0 * np.arcsin(np.sqrt(a))
    distance_m = 6367000.0 * c
    return distance_m



import csv
from os.path import join
from typing import List

import numpy as np
import pandas as pd

from aequilibrae.transit.transit_elements import Stop


def write_stops(stops: List[Stop], folder_path: str):
    data = [
        [
            stp.stop_id,
            stp.stop,
            stp.stop_name,
            stp.stop_desc,
            stp.stop_lat,
            stp.stop_lon,
            stp.zone_id,
            stp.parent_station,
        ]
        for stp in stops
    ]

    headers = ["stop_id", "stop_code", "stop_name", "stop_desc", "stop_lat", "stop_lon", "zone_id", "parent_station"]
    df = pd.DataFrame(data, columns=headers)

    df.parent_station = df.parent_station.astype(float)
    df.loc[:, "parent_station"] = np.NAN
    for fld in ["zone_id", "stop_id"]:
        df[fld] = df[fld].astype(float)
        df[fld].fillna(-99999, inplace=True)
        df[fld] = df[fld].astype(int)
        df[fld] = df[fld].astype(str)
        df.loc[df[fld] == "-99999", fld] = ""

    df.to_csv(join(folder_path, "stops.txt"), quoting=csv.QUOTE_NONNUMERIC, index=False)



from aequilibrae.paths.results import *
from aequilibrae.paths.multi_threaded_aon import MultiThreadedAoN
from aequilibrae.paths.multi_threaded_skimming import MultiThreadedNetworkSkimming
from aequilibrae.paths.network_skimming import NetworkSkimming
from aequilibrae.paths.all_or_nothing import allOrNothing
from aequilibrae.paths.assignment_paths import AssignmentPaths
from aequilibrae.paths.traffic_class import TrafficClass, TransitClass
from aequilibrae.paths.traffic_assignment import TrafficAssignment, TransitAssignment
from aequilibrae.paths.vdf import VDF
from aequilibrae.paths.graph import Graph, TransitGraph
from aequilibrae.paths.optimal_strategies import OptimalStrategies
from aequilibrae.paths.public_transport import HyperpathGenerating
from aequilibrae.paths.route_choice import RouteChoice

from aequilibrae import global_logger


try:
    from aequilibrae.paths.AoN import (
        one_to_all,
        skimming_single_origin,
        path_computation,
        update_path_trace,
    )
except ImportError as ie:
    global_logger.warning(f"Could not import procedures from the binary. {ie.args}")



from aequilibrae.paths.cython.coo_demand cimport GeneralisedCOODemand
from aequilibrae.paths.cython.route_choice_types cimport RouteVec_t
from aequilibrae.matrix.sparse_matrix cimport COO_f64_struct, COO_f32_struct

from libcpp.vector cimport vector
from libcpp.memory cimport unique_ptr, make_shared
from libcpp.unordered_set cimport unordered_set
from libcpp cimport bool

# I understand this isn't a great way to handle this but I'm rather unsure of a better method.  We need a method to
# store both float and double loads. Cython doesn't allow us to use fused types (Cython diet templates) in any manner
# that's useful. We can't store them on an object nor put them into a C++ structure, they can only be returned, used as
# a local variable, or passed to a function.
#
# Cython doesn't allow us to write a templated class, we'd have to write an actual C++ class for that. Even if we did
# that we'd still have to write a wrapper class.
cdef class LinkLoadingResults:
    cdef:
        GeneralisedCOODemand demand
        readonly object select_link_set_names
        size_t num_links

        # Number of threads
        #               * number of demand cols
        #                 |               * number of links
        vector[unique_ptr[vector[unique_ptr[vector[double]]]]] f64_link_loading_threaded

        # Number of demand cols
        #               * number of links
        vector[unique_ptr[vector[double]]] f64_link_loading

        vector[unique_ptr[vector[unique_ptr[vector[float]]]]] f32_link_loading_threaded
        vector[unique_ptr[vector[float]]] f32_link_loading

        # Select link
        # A select link set is represented by a vector of unordered AND sets, OR'd together
        # Number of select link sets
        #               * number of select link OR sets
        #                 |               * number of AND sets
        vector[unique_ptr[vector[unique_ptr[unordered_set[long long]]]]] select_link_sets

        # Number of select link sets
        #               * number of select link OR sets
        vector[unique_ptr[vector[size_t]]] select_link_set_lengths

        # Number of threads
        #               * number of select link sets
        #                 |               * number of demand cols
        #                 |                 |               * number of links
        vector[unique_ptr[vector[unique_ptr[vector[unique_ptr[vector[double]]]]]]] f64_sl_link_loading_threaded

        # Number of select link sets
        #               * number of demand cols
        #                 |               * number of links
        vector[unique_ptr[vector[unique_ptr[vector[double]]]]]  f64_sl_link_loading

        vector[unique_ptr[vector[unique_ptr[vector[unique_ptr[vector[float]]]]]]] f32_sl_link_loading_threaded
        vector[unique_ptr[vector[unique_ptr[vector[float]]]]] f32_sl_link_loading

        # Number of threads
        #               * number of select link sets
        #                 |               * number of demand cols
        vector[unique_ptr[vector[unique_ptr[vector[COO_f64_struct]]]]] f64_sl_od_matrix_threaded

        # Number of select link sets
        #               * number of demand cols
        vector[unique_ptr[vector[COO_f64_struct]]] f64_sl_od_matrix

        vector[unique_ptr[vector[unique_ptr[vector[COO_f32_struct]]]]] f32_sl_od_matrix_threaded
        vector[unique_ptr[vector[COO_f32_struct]]] f32_sl_od_matrix

        readonly object link_loading_objects
        readonly object sl_link_loading_objects
        readonly object od_matrix_objects

    cdef void link_load_single_route_set(
        LinkLoadingResults self,
        const size_t od_idx,
        const RouteVec_t &route_set,
        const vector[double] &prob_vec,
        const size_t thread_id
    ) noexcept nogil

    cdef void reduce_link_loading(LinkLoadingResults self)
    cdef object apply_generic_link_loading(
        LinkLoadingResults self,
        vector[unique_ptr[vector[double]]] &f64_link_loading,
        vector[unique_ptr[vector[float]]] &f32_link_loading,
        long long[:] compressed_id_view,
        int cores
    )

    @staticmethod
    cdef bool is_in_select_link_set(
        vector[long long] &route,
        const vector[unique_ptr[unordered_set[long long]]] &select_link_set,
        const vector[size_t] &select_link_set_lengths
    ) noexcept nogil
    cdef void sl_link_load_single_route_set(
        LinkLoadingResults self,
        const size_t od_idx,
        const RouteVec_t &route_set,
        const vector[double] &prob_vec,
        const long long origin_idx,
        const long long dest_idx,
        const size_t thread_id
    ) noexcept nogil
    cdef void reduce_sl_link_loading(LinkLoadingResults self)
    cdef void reduce_sl_od_matrix(LinkLoadingResults self)
    cdef object link_loading_to_objects(self, long long[:] compressed_id_view, int cores)
    cdef object sl_link_loading_to_objects(self, long long[:] compressed_id_view, int cores)
    cdef object sl_od_matrices_structs_to_objects(LinkLoadingResults self)
